import urllib.parse
from xml.etree.cElementTree import Element, SubElement, tostring


downloadable = {
    "bb433ea476dca6db52c1a9851c312307dda63010":
    "Cybernetix - The First Battle (1991)(Vision)(SW)[h BTTR].adf",
    "67334f9bf28b9e4ae4fb38d695689b3ff8414766":
    "Dogfight v1.1 (19xx)(R. Ling).adf",
    "ecd2f292d3f799cd67f6cf2632246836908da380":
    "Transplant.adf",
}


def find_downloadable_file(sha1):
    if sha1 in downloadable:
        return sha1_to_url(sha1)
    return None


def sha1_to_url(sha1):
    return "http://fs-uae.net/s/{0}/{1}/{2}".format(
        sha1[:2], sha1, urllib.parse.quote(downloadable[sha1].encode("UTF-8")))


def builtin_configs():
    for config_dict in configurations:
        yield create_xml_from_config_dict(config_dict)


def create_xml_from_config_dict(config_dict):
    root = Element("config")
    root.set("uuid", config_dict["uuid"])

    name_node = SubElement(root, "name")
    name_node.text = config_dict["config_name"]

    game_node = SubElement(root, "game")
    game_node.set("uuid", config_dict["parent"])
    game_name_node = SubElement(game_node, "name")
    game_name_node.text = config_dict["game_name"]
    game_platform_node = SubElement(game_node, "platform")
    game_platform_node.text = config_dict["platform"]

    options_node = SubElement(root, "options")
    for key, value in config_dict.items():
        if key in ["game_name", "platform", "config_name", "uuid", "parent"]:
            continue
        if key.startswith("file_"):
            continue
        option_node = SubElement(options_node, key)
        option_node.text = value

    for i in range(100):
        name = config_dict.get("file_{0}_name".format(i), "")
        if not name:
            break
        sha1 = config_dict.get("file_{0}_sha1".format(i), "")
        # url = config_dict.get("file_{0}_url".format(i), "")
        url = sha1_to_url(sha1)
        file_node = SubElement(root, "file")
        file_name_node = SubElement(file_node, "name")
        file_name_node.text = name
        file_sha1_node = SubElement(file_node, "sha1")
        file_sha1_node.text = sha1
        file_url_node = SubElement(file_node, "url")
        file_url_node.text = url

    name = "{0} ({1}, {2})".format(config_dict["game_name"],
                                   config_dict["platform"],
                                   config_dict["config_name"])
    print(tostring(root))
    return name, tostring(root)

configurations = [

    # http://aminet.net/game/2play/BlitzBombers.readme
    #
    # Short:        It's BLITZBOMBERS. And it's FREE.
    # Author:       redwhen@ldngedge.demon.co.uk (Red When Excited)
    # Uploader:     Big Will Riker ldngedge demon co uk (Steve Matty)
    # Type:         game/2play
    # Architecture: m68k-amigaos
    #
    # It's FINALLY here. The game some, or even a LOT people have
    # been craving for. It's FREE as well.
    #
    # It's a snapshot of the game, as it was, back in January 1996 -
    # when it was 98% complete. There are some quirks, but hey.
    #
    # Download and enjoy. Major bugs will probably get fixed.
    #
    # It requires an AGA Amiga, needs at LEAST 1.5Mb ChipRAM to run.
    # FastRAM and/or accelerator will help when using lots of CPU
    # players, or the serial link.
    #
    # http://www.ldngedge.demon.co.uk   is the place for FAQ's,
    # patches etc.
    #
    # Have fun.

    {
        "uuid": "bf1b7659-bb30-44b0-83dd-bd2c428a4214",
        "parent": "c1bcb261-adad-5a51-8f25-c5575798e30b",
        "game_name": "Cybernetix: The First Battle",
        "platform": "Amiga",
        "config_name": "Built-in, Shareware, Vision",
        "kickstart": "AROS",
        "file_0_name":
            "Cybernetix - The First Battle (1991)(Vision)(SW)[h BTTR].adf",
        "file_0_sha1": "bb433ea476dca6db52c1a9851c312307dda63010",
    },

    # Does not work with builtin AROS kickstart (WB problems)
    # {
    # "uuid":        "b52a2ddd-e314-4479-8019-575f7b7b0e7a",
    # "parent":      "b11f7f00-1fd7-519a-a0b8-ae0309053d22",
    # "game_name":   "Dogfight",
    # "platform":    "Amiga",
    # "config_name": "Built-in, Freeware, Richard Ling",
    # "kickstart":   "AROS",
    # "file_0_name": "Dogfight v1.1 (19xx)(R. Ling).adf",
    # "file_0_sha1": "67334f9bf28b9e4ae4fb38d695689b3ff8414766",
    # },

    # Does not work with builtin AROS kickstart (freezes at title screen)
    # "Downfall (Amiga, Freeware)":
    # """<config uuid="2e1998fb-41e0-4f67-9f6a-e7cddfbadffb">
    #   <name>Shareware</name>
    #   <game uuid="1b6a5c73-eaad-5601-9875-0ad3498a220a">
    #     <name>Downfall</name>
    #     <platform>Amiga</platform>
    #   </game>
    #   <options>
    #   <kickstart>AROS</kickstart>
    #   </options>
    #   <file>
    #     <name>Downfall.adf</name>
    #     <sha1>cc9ea0e8d1ae139ea5313d9b22eccc2e965744e2</sha1>
    #     <url>http://fengestad.no/sha1/cc/cc9ea0e8d1ae139ea5313d9b22eccc2e965744e2</url>
    #   </file>
    # </config>
    # """

    {
        "uuid": "ecb3da72-86c4-4921-a2c4-34f2d9290380",
        "parent": "1adf9bd3-4113-520b-973e-3caaebae369e",
        "game_name": "Transplant",
        "platform": "Amiga",
        "config_name": "Built-in, Freeware, Jumping Jack Flash",
        "kickstart": "AROS",
        "file_0_name": "Transplant.adf",
        "file_0_sha1": "ecd2f292d3f799cd67f6cf2632246836908da380",
    },
]



import atexit
import json
import os
import subprocess


def reset_gnome_extensions(extensions):
    print("reset_gnome_extensions", extensions)
    p = subprocess.Popen(
        ["gsettings", "set", "org.gnome.shell", "enabled-extensions",
         json.dumps(extensions).replace("\"", "'")])
    return p.wait() == 0


def handle_gnome_extensions():
    original_extensions = []
    try:
        p = subprocess.Popen(
            ["gsettings", "get", "org.gnome.shell", "enabled-extensions"],
            stdout=subprocess.PIPE)
        # noinspection PyUnresolvedReferences
        data = p.stdout.read().decode("UTF-8", errors="replace")
        if p.wait() == 0:
            print("enabled-extensions:")
            print(data)
            if data.startswith("@as "):
                data = data[4:]
            original_extensions = json.loads(data.replace("'", "\""))
            print(original_extensions)
    except FileNotFoundError:
        print("no gsettings program found")
        return
    except ValueError as e:
        print(repr(e))
        print("could not decode enabled-exceptions")
        return

    new_extensions = original_extensions.copy()
    if "window-list@gnome-shell-extensions.gcampax.github.com" \
            in original_extensions:
        new_extensions.remove(
            "window-list@gnome-shell-extensions.gcampax.github.com")
    if "hidetopbar@mathieu.bidon.ca" not in original_extensions:
        new_extensions.append("hidetopbar@mathieu.bidon.ca")

    if new_extensions != original_extensions:
        p = subprocess.Popen(
            ["gsettings", "set", "org.gnome.shell", "enabled-extensions",
             json.dumps(new_extensions).replace("\"", "'")])
        if p.wait() == 0:
            print("atexit.register reset_gnome_extensions")
            atexit.register(reset_gnome_extensions, original_extensions)

    # sys.exit(1)


def running_in_gnome_3():
    value = os.environ.get("XDG_CURRENT_DESKTOP", "").upper()
    if value:
        return value == "GNOME"
    value = os.environ.get("DESKTOP_SESSION", "").upper()
    if value:
        return value == "GNOME"
    return False



"""Static type checking helpers"""

from abc import ABCMeta, abstractmethod, abstractproperty
import collections
import inspect
import sys
import re


__all__ = [
    # Type system related
    'AbstractGeneric',
    'AbstractGenericMeta',
    'Any',
    'AnyStr',
    'Dict',
    'Function',
    'Generic',
    'GenericMeta',
    'IO',
    'List',
    'Match',
    'NamedTuple',
    'Pattern',
    'Protocol',
    'Set',
    'Tuple',
    'Undefined',
    'Union',
    'cast',
    'forwardref',
    'overload',
    'typevar',
    # Protocols and abstract base classes
    'Container',
    'Iterable',
    'Iterator',
    'Sequence',
    'Sized',
    'AbstractSet',
    'Mapping',
    'BinaryIO',
    'TextIO',
]


def builtinclass(cls):
    """Mark a class as a built-in/extension class for type checking."""
    return cls


def ducktype(type):
    """Return a duck type declaration decorator.

    The decorator only affects type checking.
    """
    def decorator(cls):
        return cls
    return decorator


def disjointclass(type):
    """Return a disjoint class declaration decorator.

    The decorator only affects type checking.
    """
    def decorator(cls):
        return cls
    return decorator


class GenericMeta(type):
    """Metaclass for generic classes that support indexing by types."""

    def __getitem__(self, args):
        # Just ignore args; they are for compile-time checks only.
        return self


class Generic(metaclass=GenericMeta):
    """Base class for generic classes."""


class AbstractGenericMeta(ABCMeta):
    """Metaclass for abstract generic classes that support type indexing.

    This is used for both protocols and ordinary abstract classes.
    """

    def __new__(mcls, name, bases, namespace):
        cls = super().__new__(mcls, name, bases, namespace)
        # 'Protocol' must be an explicit base class in order for a class to
        # be a protocol.
        cls._is_protocol = name == 'Protocol' or Protocol in bases
        return cls

    def __getitem__(self, args):
        # Just ignore args; they are for compile-time checks only.
        return self


class Protocol(metaclass=AbstractGenericMeta):
    """Base class for protocol classes."""

    @classmethod
    def __subclasshook__(cls, c):
        if not cls._is_protocol:
            # No structural checks since this isn't a protocol.
            return NotImplemented

        if cls is Protocol:
            # Every class is a subclass of the empty protocol.
            return True

        # Find all attributes defined in the protocol.
        attrs = cls._get_protocol_attrs()

        for attr in attrs:
            if not any(attr in d.__dict__ for d in c.__mro__):
                return NotImplemented
        return True

    @classmethod
    def _get_protocol_attrs(cls):
        # Get all Protocol base classes.
        protocol_bases = []
        for c in cls.__mro__:
            if getattr(c, '_is_protocol', False) and c.__name__ != 'Protocol':
                protocol_bases.append(c)

        # Get attributes included in protocol.
        attrs = set()
        for base in protocol_bases:
            for attr in base.__dict__.keys():
                # Include attributes not defined in any non-protocol bases.
                for c in cls.__mro__:
                    if (c is not base and attr in c.__dict__ and
                            not getattr(c, '_is_protocol', False)):
                        break
                else:
                    if (not attr.startswith('_abc_') and
                        attr != '__abstractmethods__' and
                        attr != '_is_protocol' and
                        attr != '__dict__' and
                        attr != '_get_protocol_attrs' and
                        attr != '__module__'):
                        attrs.add(attr)

        return attrs


class AbstractGeneric(metaclass=AbstractGenericMeta):
    """Base class for abstract generic classes."""


class TypeAlias:
    """Class for defining generic aliases for library types."""

    def __init__(self, target_type):
        self.target_type = target_type

    def __getitem__(self, typeargs):
        return self.target_type


Traceback = object()  # TODO proper type object


# Define aliases for built-in types that support indexing.
List = TypeAlias(list)
Dict = TypeAlias(dict)
Set = TypeAlias(set)
Tuple = TypeAlias(tuple)
Function = TypeAlias(callable)
Pattern = TypeAlias(type(re.compile('')))
Match = TypeAlias(type(re.match('', '')))


def union(x): return x


Union = TypeAlias(union)


def NamedTuple(typename, fields):
    return collections.namedtuple(typename,
                                  (name for name, type in fields))


class typevar:
    def __init__(self, name, *, values=None):
        self.name = name
        self.values = values


# Predefined type variables.
AnyStr = typevar('AnyStr', values=(str, bytes))


class forwardref:
    def __init__(self, name):
        self.name = name


def Any(x):
    """The Any type; can also be used to cast a value to type Any."""
    return x

def cast(type, object):
    """Cast a value to a type.

    This only affects static checking; simply return object at runtime.
    """
    return object


def overload(func):
    """Function decorator for defining overloaded functions."""
    frame = sys._getframe(1)
    locals = frame.f_locals
    # See if there is a previous overload variant available.  Also verify
    # that the existing function really is overloaded: otherwise, replace
    # the definition.  The latter is actually important if we want to reload
    # a library module such as genericpath with a custom one that uses
    # overloading in the implementation.
    if func.__name__ in locals and hasattr(locals[func.__name__], 'dispatch'):
        orig_func = locals[func.__name__]

        def wrapper(*args, **kwargs):
            ret, ok = orig_func.dispatch(*args, **kwargs)
            if ok:
                return ret
            return func(*args, **kwargs)
        wrapper.isoverload = True
        wrapper.dispatch = make_dispatcher(func, orig_func.dispatch)
        wrapper.next = orig_func
        wrapper.__name__ = func.__name__
        if hasattr(func, '__isabstractmethod__'):
            # Note that we can't reliably check that abstractmethod is
            # used consistently across overload variants, so we let a
            # static checker do it.
            wrapper.__isabstractmethod__ = func.__isabstractmethod__
        return wrapper
    else:
        # Return the initial overload variant.
        func.isoverload = True
        func.dispatch = make_dispatcher(func)
        func.next = None
        return func


def is_erased_type(t):
    return t is Any or isinstance(t, typevar)


def make_dispatcher(func, previous=None):
    """Create argument dispatcher for an overloaded function.

    Also handle chaining of multiple overload variants.
    """
    (args, varargs, varkw, defaults,
     kwonlyargs, kwonlydefaults, annotations) = inspect.getfullargspec(func)

    argtypes = []
    for arg in args:
        ann = annotations.get(arg)
        if isinstance(ann, forwardref):
            ann = ann.name
        if is_erased_type(ann):
            ann = None
        elif isinstance(ann, str):
            # The annotation is a string => evaluate it lazily when the
            # overloaded function is first called.
            frame = sys._getframe(2)
            t = [None]
            ann_str = ann
            def check(x):
                if not t[0]:
                    # Evaluate string in the context of the overload caller.
                    t[0] = eval(ann_str, frame.f_globals, frame.f_locals)
                    if is_erased_type(t[0]):
                        # Anything goes.
                        t[0] = object
                if isinstance(t[0], type):
                    return isinstance(x, t[0])
                else:
                    return t[0](x)
            ann = check
        argtypes.append(ann)

    maxargs = len(argtypes)
    minargs = maxargs
    if defaults:
        minargs = len(argtypes) - len(defaults)

    def dispatch(*args, **kwargs):
        if previous:
            ret, ok = previous(*args, **kwargs)
            if ok:
                return ret, ok

        nargs = len(args)
        if nargs < minargs or nargs > maxargs:
            # Invalid argument count.
            return None, False

        for i in range(nargs):
            argtype = argtypes[i]
            if argtype:
                if isinstance(argtype, type):
                    if not isinstance(args[i], argtype):
                        break
                else:
                    if not argtype(args[i]):
                        break
        else:
            return func(*args, **kwargs), True
        return None, False
    return dispatch


class Undefined:
    """Class that represents an undefined value with a specified type.

    At runtime the name Undefined is bound to an instance of this
    class.  The intent is that any operation on an Undefined object
    raises an exception, including use in a boolean context.  Some
    operations cannot be disallowed: Undefined can be used as an
    operand of 'is', and it can be assigned to variables and stored in
    containers.

    'Undefined' makes it possible to declare the static type of a
    variable even if there is no useful default value to initialize it
    with:

      from typing import Undefined
      x = Undefined(int)
      y = Undefined # type: int

    The latter form can be used if efficiency is of utmost importance,
    since it saves a call operation and potentially additional
    operations needed to evaluate a type expression.  Undefined(x)
    just evaluates to Undefined, ignoring the argument value.
    """

    def __repr__(self):
        return '<typing.Undefined>'

    def __setattr__(self, attr, value):
        raise AttributeError("'Undefined' object has no attribute '%s'" % attr)

    def __eq__(self, other):
        raise TypeError("'Undefined' object cannot be compared")

    def __call__(self, type):
        return self

    def __bool__(self):
        raise TypeError("'Undefined' object is not valid as a boolean")


Undefined = Undefined()


# Abstract classes


T = typevar('T')
KT = typevar('KT')
VT = typevar('VT')


class SupportsInt(Protocol):
    @abstractmethod
    def __int__(self) -> int: pass


class SupportsFloat(Protocol):
    @abstractmethod
    def __float__(self) -> float: pass


class SupportsAbs(Protocol[T]):
    @abstractmethod
    def __abs__(self) -> T: pass


class SupportsRound(Protocol[T]):
    @abstractmethod
    def __round__(self, ndigits: int = 0) -> T: pass


class Reversible(Protocol[T]):
    @abstractmethod
    def __reversed__(self) -> 'Iterator[T]': pass


class Sized(Protocol):
    @abstractmethod
    def __len__(self) -> int: pass


class Container(Protocol[T]):
    @abstractmethod
    def __contains__(self, x) -> bool: pass


class Iterable(Protocol[T]):
    @abstractmethod
    def __iter__(self) -> 'Iterator[T]': pass


class Iterator(Iterable[T], Protocol[T]):
    @abstractmethod
    def __next__(self) -> T: pass


class Sequence(Sized, Iterable[T], Container[T], AbstractGeneric[T]):
    @abstractmethod
    @overload
    def __getitem__(self, i: int) -> T: pass

    @abstractmethod
    @overload
    def __getitem__(self, s: slice) -> 'Sequence[T]': pass

    @abstractmethod
    def __reversed__(self, s: slice) -> Iterator[T]: pass

    @abstractmethod
    def index(self, x) -> int: pass

    @abstractmethod
    def count(self, x) -> int: pass


for t in list, tuple, str, bytes, range:
    Sequence.register(t)


class AbstractSet(Sized, Iterable[T], AbstractGeneric[T]):
    @abstractmethod
    def __contains__(self, x: object) -> bool: pass
    @abstractmethod
    def __and__(self, s: 'AbstractSet[T]') -> 'AbstractSet[T]': pass
    @abstractmethod
    def __or__(self, s: 'AbstractSet[T]') -> 'AbstractSet[T]': pass
    @abstractmethod
    def __sub__(self, s: 'AbstractSet[T]') -> 'AbstractSet[T]': pass
    @abstractmethod
    def __xor__(self, s: 'AbstractSet[T]') -> 'AbstractSet[T]': pass
    @abstractmethod
    def isdisjoint(self, s: 'AbstractSet[T]') -> bool: pass


for t in set, frozenset, type({}.keys()), type({}.items()):
    AbstractSet.register(t)


class Mapping(Sized, Iterable[KT], AbstractGeneric[KT, VT]):
    @abstractmethod
    def __getitem__(self, k: KT) -> VT: pass
    @abstractmethod
    def __setitem__(self, k: KT, v: VT) -> None: pass
    @abstractmethod
    def __delitem__(self, v: KT) -> None: pass
    @abstractmethod
    def __contains__(self, o: object) -> bool: pass

    @abstractmethod
    def clear(self) -> None: pass
    @abstractmethod
    def copy(self) -> 'Mapping[KT, VT]': pass
    @overload
    @abstractmethod
    def get(self, k: KT) -> VT: pass
    @overload
    @abstractmethod
    def get(self, k: KT, default: VT) -> VT: pass
    @overload
    @abstractmethod
    def pop(self, k: KT) -> VT: pass
    @overload
    @abstractmethod
    def pop(self, k: KT, default: VT) -> VT: pass
    @abstractmethod
    def popitem(self) -> Tuple[KT, VT]: pass
    @overload
    @abstractmethod
    def setdefault(self, k: KT) -> VT: pass
    @overload
    @abstractmethod
    def setdefault(self, k: KT, default: VT) -> VT: pass

    @overload
    @abstractmethod
    def update(self, m: 'Mapping[KT, VT]') -> None: pass
    @overload
    @abstractmethod
    def update(self, m: Iterable[Tuple[KT, VT]]) -> None: pass

    @abstractmethod
    def keys(self) -> AbstractSet[KT]: pass
    @abstractmethod
    def values(self) -> AbstractSet[VT]: pass
    @abstractmethod
    def items(self) -> AbstractSet[Tuple[KT, VT]]: pass


# TODO Consider more types: os.environ, etc. However, these add dependencies.
Mapping.register(dict)


# Note that the BinaryIO and TextIO classes must be in sync with typing module
# stubs.


class IO(AbstractGeneric[AnyStr]):
    @abstractproperty
    def mode(self) -> str: pass
    @abstractproperty
    def name(self) -> str: pass
    @abstractmethod
    def close(self) -> None: pass
    @abstractmethod
    def closed(self) -> bool: pass
    @abstractmethod
    def fileno(self) -> int: pass
    @abstractmethod
    def flush(self) -> None: pass
    @abstractmethod
    def isatty(self) -> bool: pass
    @abstractmethod
    def read(self, n: int = -1) -> AnyStr: pass
    @abstractmethod
    def readable(self) -> bool: pass
    @abstractmethod
    def readline(self, limit: int = -1) -> AnyStr: pass
    @abstractmethod
    def readlines(self, hint: int = -1) -> List[AnyStr]: pass
    @abstractmethod
    def seek(self, offset: int, whence: int = 0) -> int: pass
    @abstractmethod
    def seekable(self) -> bool: pass
    @abstractmethod
    def tell(self) -> int: pass
    @abstractmethod
    def truncate(self, size: int = None) -> int: pass
    @abstractmethod
    def writable(self) -> bool: pass
    @abstractmethod
    def write(self, s: AnyStr) -> int: pass
    @abstractmethod
    def writelines(self, lines: List[AnyStr]) -> None: pass

    @abstractmethod
    def __enter__(self) -> 'IO[AnyStr]': pass
    @abstractmethod
    def __exit__(self, type, value, traceback) -> None: pass


class BinaryIO(IO[bytes]):
    @overload
    @abstractmethod
    def write(self, s: bytes) -> int: pass
    @overload
    @abstractmethod
    def write(self, s: bytearray) -> int: pass

    @abstractmethod
    def __enter__(self) -> 'BinaryIO': pass


class TextIO(IO[str]):
    @abstractproperty
    def buffer(self) -> BinaryIO: pass
    @abstractproperty
    def encoding(self) -> str: pass
    @abstractproperty
    def errors(self) -> str: pass
    @abstractproperty
    def line_buffering(self) -> bool: pass
    @abstractproperty
    def newlines(self) -> Any: pass
    @abstractmethod
    def __enter__(self) -> 'TextIO': pass


# TODO Register IO/TextIO/BinaryIO as the base class of file-like types.


del t



'''OpenGL extension VERSION.GLES3_3_0

This module customises the behaviour of the 
OpenGL.raw.GLES3.VERSION.GLES3_3_0 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GLES3_3_0.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper

import ctypes
from OpenGL.raw.GLES3 import _types
from OpenGL.raw.GLES3.VERSION.GLES3_3_0 import *
from OpenGL.raw.GLES3.VERSION.GLES3_3_0 import _EXTENSION_NAME

def glInitGles330VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION

glGenVertexArrays = wrapper.wrapper(glGenVertexArrays).setOutput(
    'arrays', lambda n: (n,), 'n', arrayType = arrays.GLuintArray,
    orPassIn=True,
)



from fs_uae_launcher.ui.behaviors.amigaenablebehavior import AmigaEnableBehavior
from fs_uae_launcher.ui.behaviors.configbehavior import ConfigBehavior
from fs_uae_launcher.ui.config.ConfigCheckBox import ConfigCheckBox
from fsbc.util import unused
import fsui as fsui
from ...Config import Config
from fsgs.amiga.Amiga import Amiga
from ...FloppyManager import FloppyManager
from ...CDManager import CDManager
from ...I18N import gettext
# from .ConfigCheckBox import ConfigCheckBox


class ModelGroup(fsui.Group):

    # FIXME: remove with_more_button=True
    def __init__(self, parent, with_more_button=True):
        unused(with_more_button)
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.model_ids = [
            x["id"] for x in Amiga.models if "/" not in x["id"]]
        self.model_titles = [
            x["title"] for x in Amiga.models if "/" not in x["id"]]

        self.sub_model_ids = []
        self.sub_model_titles = []
        self.sub_model_updating = False

        self.model_choice = fsui.Choice(self, self.model_titles)
        AmigaEnableBehavior(self.model_choice)
        self.sub_model_choice = fsui.Choice(self, self.sub_model_titles)
        AmigaEnableBehavior(self.sub_model_choice)
        self.accuracy_label = fsui.Label(self, gettext("Accuracy:"))
        self.accuracy_choice = fsui.Choice(self, [
            gettext("High"),
            gettext("Medium"),
            gettext("Low")])
        AmigaEnableBehavior(self.accuracy_choice)
        self.ntsc_checkbox = ConfigCheckBox(self, "NTSC", "ntsc_mode")
        AmigaEnableBehavior(self.ntsc_checkbox)

        # if fs_uae_launcher.ui.get_screen_size()[1] > 768:
        # self.layout.add(heading_label, margin=10)
        # self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        heading_label = fsui.HeadingLabel(self, gettext("Amiga Model"))
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(10)
        hori_layout.add(self.ntsc_checkbox, expand=False,
                        margin_left=10, margin_right=10)
        hori_layout.add_spacer(0, expand=True)

        hori_layout.add(self.accuracy_label, margin_right=10)
        hori_layout.add(self.accuracy_choice, margin_right=10)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        hori_layout.add(self.model_choice, expand=False, margin=10)
        hori_layout.add(self.sub_model_choice, expand=True, margin=10)

        ConfigBehavior(self, ["accuracy", "amiga_model"])

        self.model_choice.on_change = self.on_model_change
        self.sub_model_choice.on_change = self.on_sub_model_change
        self.accuracy_choice.on_change = self.on_accuracy_change

    def on_model_change(self):
        print("ModelGroup.on_model_change\n")
        index = self.model_choice.get_index()
        model = self.model_ids[index]
        if model == "A500":
            # The default model (A500) can be specified with the empty string
            model = ""
        Config.set("amiga_model", model)
        # Config.update_kickstart()
        # if Amiga.is_cd_based(Config):
        #     FloppyManager.clear_all()
        # else:
        #     CDManager.clear_all()

    def on_sub_model_change(self):
        print("ModelGroup.on_sub_model_change\n")
        if self.sub_model_updating:
            print("sub model list is currently updating")
            return
        index = self.sub_model_choice.get_index()
        # if index == 0:
        #     # The default model (A500) can be specified with the empty string
        #     model = ""
        # else:
        model = self.model_ids[self.model_choice.get_index()]
        sub_model = self.sub_model_ids[index]
        if sub_model:
            Config.set("amiga_model", model + "/" + sub_model)
        else:
            Config.set("amiga_model", model)

        if Amiga.is_cd_based(Config):
            FloppyManager.clear_all()
        else:
            CDManager.clear_all()

    def on_accuracy_change(self):
        index = self.accuracy_choice.get_index()
        if index == 0:
            Config.set("accuracy", "")
        else:
            Config.set("accuracy", str(1 - index))

    def update_sub_models(self, model_id, sub_model_id):
        sub_model_index = 0
        model_id_s = model_id + "/"
        self.sub_model_ids.clear()
        self.sub_model_titles.clear()

        for i, config in enumerate(Amiga.models):
            if config["id"] == model_id:
                self.sub_model_ids.append("")
                self.sub_model_titles.append(config["subtitle"])
            elif config["id"].startswith(model_id_s):
                self.sub_model_ids.append(config["id"].split("/", 1)[1])
                self.sub_model_titles.append(config["subtitle"])
            else:
                continue
            if sub_model_id == self.sub_model_ids[-1]:
                sub_model_index = len(self.sub_model_ids) - 1

        self.sub_model_choice.clear()
        for title in self.sub_model_titles:
            self.sub_model_choice.add_item(title)
        self.sub_model_choice.enable(len(self.sub_model_ids) > 1)
        return sub_model_index

    def on_amiga_model_config(self, value):
        if value == "":
            value = "A500"

        if "/" in value:
            model_id, sub_model_id = value.split("/", 1)
        else:
            model_id = value
            sub_model_id = ""

        model_index = 0
        sub_model_index = 0
        self.sub_model_updating = True
        for i, config in enumerate(Amiga.models_config):
            if config == value:
                # self.model_choice.set_index(i)
                # find main model index
                model_index = self.model_ids.index(model_id)
                sub_model_index = self.update_sub_models(model_id, sub_model_id)
                # model_index = i
                break
        # else:
        #    print("FIXME: could not set model")
        self.model_choice.set_index(model_index)
        self.sub_model_choice.set_index(sub_model_index)
        self.sub_model_updating = False

    def on_accuracy_config(self, value):
        if not value:
            index = 0
        else:
            index = 1 - int(value)
        self.accuracy_choice.set_index(index)



from fsui.qt import QSpinBox
from .Widget import Widget


class SpinCtrl(Widget):

    def __init__(self, parent, min_value, max_value, initial_value):
        self._widget = QSpinBox(parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self._widget.setRange(min_value, max_value)
        self._widget.setValue(initial_value)
        self._widget.valueChanged.connect(self.__value_changed)

    def get_value(self):
        return self._widget.value()

    def set_value(self, value):
        self._widget.setValue(value)

    def __value_changed(self, value):
        self.on_change()

    def on_change(self):
        pass



import os
import struct
import hashlib

from fsbc.system import windows
from fsbc.util import memoize
from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.input.mapper import InputMapper
from fsgs.runner import GameRunner


# noinspection PyAttributeOutsideInit
class MednafenRunner(GameRunner):

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def prepare(self):
        # self.temp_home = self.create_temp_dir("mednafen-home")
        with open(self.mednafen_cfg_path(), "w", encoding="UTF-8") as f:
            self.mednafen_configure(f)

    def run(self):
        # self.env["HOME"] = self.temp_home.path

        # executable = None
        # if windows:
        #     found_executable = self.find_emulator_executable(
        #         "fs-mednafen/mednafen")
        #     if not found_executable:
        #         raise Exception("could not find mednafen.exe")
        #     dir_path = os.path.dirname(found_executable)
        #     for name in os.listdir(dir_path):
        #         src = os.path.join(dir_path, name)
        #         dst = os.path.join(self.temp_home.path, name)
        #         shutil.copyfile(src, dst)
        #     executable = os.path.join(
        #         self.temp_home.path, "mednafen.exe")
        #
        # return self.start_emulator(
        #     "fs-mednafen/mednafen", args=self.args, env_vars=self.env,
        #     executable=executable)

        return self.start_emulator_from_plugin_resource(
            "fs-mednafen", args=self.args, env_vars=self.env)

    def finish(self):
        pass

    def get_supported_filters(self):
        supported = [
            {
                "name": "2x",
                "special": "nn2x",
            }, {
                "name": "none",
                "special": "none",
            }, {
                "name": "scale2x",
                "special": "scale2x",
            }, {
                "name": "hq2x",
                "special": "hq2x",
            }
        ]
        return supported

    def mednafen_input_mapping(self, port):
        raise NotImplementedError()

    def mednafen_system_prefix(self):
        raise NotImplementedError()

    def mednafen_video_size(self):
        return 0, 0

    def mednafen_viewport(self):
        viewport = self.config["viewport"]
        if viewport:
            src, dst = viewport.split("=")
            src = src.strip()
            dst = dst.strip()
            src_x, src_y, src_w, src_h = [int(v) for v in src.split(" ")]
            dst_x, dst_y, dst_w, dst_h = [int(v) for v in dst.split(" ")]
            return (src_x, src_y, src_w, src_h), (dst_x, dst_y, dst_w, dst_h)
        return None, None

    def mednafen_configure(self, f):
        self.args = []
        self.env = {}
        pfx = self.mednafen_system_prefix()

        screen_w, screen_h = self.screen_size()

        dest_w, dest_h = screen_w, screen_h
        # a_ratio = self.force_aspect_ratio()
        a_ratio = None
        # game_w, game_h = self.mednafen_video_size()
        src, dst = self.mednafen_viewport()
        if src is None:
            game_w, game_h = self.mednafen_video_size()
        else:
            game_w, game_h = dst[2], dst[3]

        self.args.extend(["-{0}.xres".format(pfx), str(screen_w)])
        self.args.extend(["-{0}.yres".format(pfx), str(screen_h)])

        if game_w and game_h:
            real_game_size = (game_w, game_h)
            if self.use_stretching():
                x_scale = dest_w / game_w
                y_scale = dest_h / game_h

                if src is not None and dst is not None:
                    x_offset = dest_w * (
                        (src[2] - dst[2]) / 2 - dst[0]) / dst[2]
                    y_offset = dest_h * (
                        (src[3] - dst[3]) / 2 - dst[1]) / dst[3]
                    x_offset = int(round(x_offset))
                    y_offset = int(round(y_offset))
                    self.env["FSGS_OFFSET"] = "{0},{1}".format(
                        x_offset, y_offset)
                    print("FSGS_OFFSET", self.env["FSGS_OFFSET"])
            else:
                if a_ratio:
                    print("Forcing aspect ratio %f", a_ratio)
                    game_h = game_w / a_ratio
                x_scale = min(dest_w / game_w, dest_h / game_h)
                if a_ratio:
                    y_scale = (x_scale * real_game_size[0] /
                               real_game_size[1] / a_ratio)
                else:
                    y_scale = x_scale
            print("Fullscreen scale factors: %f %f", x_scale, y_scale)
            self.args.extend(["-%s.xscalefs" % pfx, str(x_scale)])
            self.args.extend(["-%s.yscalefs" % pfx, str(y_scale)])
            self.args.extend(["-%s.stretch" % pfx, "0"])
        else:
            if self.use_stretching():
                self.args.extend(["-%s.stretch" % pfx, "full"])
            else:
                self.args.extend(["-%s.stretch" % pfx, "aspect"])

        if self.use_fullscreen():
            # self.args.extend(["-fs", "1"])
            # self.args.extend(["-%s.scanlines" % pfx,
            #         str(self.get_scanlines_setting())])
            pass
        else:
            # self.args.extend(["-fs", "0"])
            self.args.extend(["-%s.xscale" % pfx, "2"])
            self.args.extend(["-%s.yscale" % pfx, "2"])

            # self.args.extend(["-%s.scanlines" % pfx, "0"])
            # if gamew < 200:
            #     self.args.extend(["-%s.xscale" % pfx, "4"])
            #     self.args.extend(["-%s.yscale" % pfx, "4"])
            # else:
            #     self.args.extend(["-%s.xscale" % pfx, "3"])
            #     self.args.extend(["-%s.yscale" % pfx, "3"])
            # self.args.extend(["-%s.videoip" % pfx, "0"])
            #  FIXME:
            # self.options.smooth = False
            # self.options.filter = '2x'

        if self.use_doubling():
            self.args.extend(["-%s.special" % pfx, "nn2x"])

        self.args.extend(self.mednafen_extra_graphics_options())

        # filter_data = self.configure_filter()
        # if filter_data["name"] == "ntsc":
        #     self.args.extend(["-{0}.ntscblitter".format(pfx), "1"])
        #     self.args.extend(["-{0}.ntsc.preset".format(pfx), "composite"])
        #     self.args.extend(["-{0}.ntsc.saturation".format(pfx), "0.5"])
        # else:
        #     self.args.extend(
        #         ["-{0}.special".format(pfx), filter_data["special"]])

        if self.use_smoothing():
            self.args.extend(["-%s.videoip" % pfx, "1"])

        # self.args.extend(["-%s.special" % pfx,
        # self.mednafen_special_filter()])

        if self.configure_vsync():
            self.args.extend(["-glvsync", "1"])
        else:
            self.args.extend(["-glvsync", "0"])

        if self.config.get("audio_driver", "") in ["sdl", "pulseaudio"]:
            # Mednafen does not support pulseaudio directly, but using the
            # sdl driver will "often" result in pulseaudio being used
            # indirectly
            self.args.extend(["-sound.driver", "sdl"])

        print("\n" + "-" * 79 + "\n" + "CONFIGURE PORTS")

        for i, port in enumerate(self.ports):
            input_mapping = self.mednafen_input_mapping(i)
            mapper = MednafenInputMapper(port, input_mapping)
            keys = {}
            for key, value in mapper.items():
                keys.setdefault(key, []).append(value)
            for key, values in keys.items():
                print(repr(key), repr(values))
                f.write("{key} {value}\n".format(
                        key=key, value="~".join(values)))

        print("\n" + "-" * 79 + "\n" + "CONFIGURE DIRS")

        state_dir = self.emulator_state_dir("mednafen")
        self.args.extend(["-path_sav", state_dir])
        self.args.extend(["-path_state", state_dir])
        self.args.extend(["-filesys.fname_state", "%M%X"])
        self.args.extend(["-filesys.fname_sav", "%M%x"])

        # docdir = pyapp.user.documents_dir()
        self.doc_dir = self.create_temp_dir("mednafen-docs")
        self.args.extend([
            "-path_movie", self.doc_dir.path,
            "-path_cheat", self.doc_dir.path,
            "-path_palette", self.home.path])

        self.args.extend(["-path_snap", self.screenshots_dir()])
        self.args.extend(["-filesys.fname_snap",
                          "{0}-%p.%x".format(self.screenshots_name())])

        self.args.append(self.get_game_file())
        self.mednafen_post_configure()

    def set_mednafen_input_order(self):
        if windows:
            self.input_device_order = 'DINPUT8'
        self.input_mapping_multiple = False

    def mednafen_extra_graphics_options(self):
        return []

    def mednafen_post_configure(self):
        # can be overriden by subclasses
        pass

    # def mednafen_refresh_rate(self):
    #     return 0.0
    #
    # def get_game_refresh_rate(self):
    #     # can be overriden by subclasses
    #     return self.mednafen_refresh_rate()

    def mednafen_cfg_path(self):
        if not os.path.exists(os.path.join(self.home.path, ".mednafen")):
            os.makedirs(os.path.join(self.home.path, ".mednafen"))
        return os.path.join(self.home.path, ".mednafen", "mednafen-09x.cfg")

        # config_path = os.path.join(os.environ['HOME'], '.mednafen')

        # the SDL version (even on Windows) seems to read HOME, so
        # use environment instead of pyapp.user.home_dir

        # try:
        #     config_path = os.path.join(os.environ['HOME'], '.mednafen')
        # except KeyError:
        #     config_path = os.path.join(pyapp.user.home_dir(), '.mednafen')
        # self.config_temp = self.create_temp_file(".mednafen")
        #  self.config_temp = self.create_temp_file(".mednafen")
        #  config_path = os.path.join(self.context.temp.dir('mednafen'),
        #  '.mednafen')
        # if not os.path.isdir(config_path):
        #     os.makedirs(config_path)
        # return os.path.join(config_path, "mednafen.cfg")
        # return self.config_temp.path

        # elif fs.windows:
        #     config_path = os.path.join(pyapp.fs.data_dir_user_app(),
        #             "mednafen")
        #     os.putenv("HOME", str_path_path(config_path))
        #     if not os.path.exists(config_path):
        #         os.makedirs(config_path)
        #     config_path = os.path.join(config_path, ".mednafen")
        #     if not os.path.exists(config_path):
        #         os.makedirs(config_path)
        #     return os.path.join(config_path, "mednafen.cfg")
        # elif fs.macosx:
        #     config_path = os.path.join(pyapp.user.home_dir(), "Library",
        #             "Application Support", "ZSNES")
        #     if not os.path.isdir(config_path):
        #         os.makedirs(config_path)
        #     return os.path.join(config_path, "zsnesl.cfg")

    # def get_joystick_unique_ids(self, controllers):
    #     unique_ids = {}
    #     for i in range(len(controllers)):
    #         controller = controllers[i]
    #         unique_id = self.get_joystick_unique_id(controller)
    #         plusplus = unique_ids.values().count(unique_id)
    #         unique_id += plusplus
    #         unique_ids[controller.id] = unique_id
    #     return unique_ids

    def is_pal(self):
        # return self.config.get("ntsc_mode") != "1"
        # return False
        refresh_rate = self.get_game_refresh_rate()
        if refresh_rate:
            return int(round(refresh_rate)) == 50


class MednafenInputMapper(InputMapper):

    def __init__(self, input, mapping):
        InputMapper.__init__(self, input, mapping)
        helper = EnumerateHelper()
        helper.init()
        seen_ids = set()
        self.id_map = {}
        for device in helper.devices:
            uid = self._get_unique_id(device)
            while uid in seen_ids:
                uid += 1
            seen_ids.add(uid)
            self.id_map[device.id] = uid
        print("MednafenInputMapper device map")
        for id, uid in self.id_map.items():
            print(uid, id)

    def axis(self, axis, positive):
        if positive:
            offset = 0x8000
        else:
            offset = 0xc000
        joystick_id = self.get_unique_id(self.device, self.device.id)
        return "joystick {0:x} {1:08x}".format(
            joystick_id, axis + offset)

    def hat(self, _, direction):
        offset = {
            "left": 8,
            "right": 2,
            "up": 1,
            "down": 4,
        }[direction]
        joystick_id = self.get_unique_id(self.device, self.device.id)
        return "joystick {0:x} {1:08x}".format(
            joystick_id, 0x2000 + offset)

    def button(self, button):
        joystick_id = self.get_unique_id(self.device, self.device.id)
        return "joystick {0:x} {1:08x}".format(
            joystick_id, int(button))

    def key(self, key):
        # FIXME: Need other key codes on Windows ... ?
        return "keyboard {0}".format(key.sdl_code)

    @memoize
    def get_unique_id(self, device, device_id):
        # unique_id = self._get_unique_id(device)
        # print("joystick unique id:", unique_id)
        # return unique_id
        return self.id_map[device.id]

    @memoize
    def _get_unique_id(self, device):
        print("get_unique_id for", device.id)

        # Implemented the algorithm in mednafen
        # was src/drivers/joystick.cpp:GetJoystickUniqueID
        # now src/drivers/Joystick.cpp:CalcOldStyleID

        m = hashlib.md5()
        print(device.axes, device.balls, device.hats, device.buttons)
        buffer = struct.pack("iiii", device.axes, device.balls,
                             device.hats, device.buttons)

        m.update(buffer)
        digest = m.digest()
        ret = 0
        for x in range(16):
            # ret ^= ord(digest[x]) << ((x & 7) * 8)
            ret ^= digest[x] << ((x & 7) * 8)
        return ret



import os
from fsbc.Application import app
from fsgs.runner import GameRunner
from fsgs.amiga.FSUAE import FSUAE
from fsgs.amiga.LaunchHandler import LaunchHandler


class AmigaRunner(GameRunner):

    JOYSTICK = {
        "type": "joystick",
        "description": "Joystick",
        "mapping_name": "amiga",
    }

    MOUSE = {
        "type": "mouse",
        "description": "Mouse",
        "mapping_name": "amiga_mouse",  # FIXME ...
    }

    PORTS = [
        {
            "description": "Joystick Port",
            "types": [JOYSTICK]
        }, {
            "description": "Mouse Port",
            "types": [MOUSE, JOYSTICK]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.temp_config_file = None
        self.launch_handler = None

    # def init(self):
    #     print("AmigaGameHandler.init")
    #
    #     loader = ValueConfigLoader(self.fsgs.game.variant_uuid)
    #     loader.load_values(self.fsgs.game.config)
    #
    #     config = defaultdict(six.text_type)
    #
    #     for key, value in six.iteritems(app.settings.values):
    #         # FIXME: re-enable this check?
    #         # if key in Config.config_keys:
    #         #     print("... ignoring config key from settings:", key)
    #         #     continue
    #         config[key] = value
    #
    #     original_config = loader.get_config()
    #     for key, value in six.iteritems(original_config):
    #         config[key] = value
    #
    #     # make sure FS-UAE does not load other config files (Host.fs-uae)
    #     config["end_config"] = "1"
    #
    #     self.port1_option = InputOption(
    #         1, self.fsgs.game.config.get("port1", "joystick"))
    #     #self.options.append(self.port1_option)
    #
    #     self.port0_option = InputOption(
    #         0, self.fsgs.game.config.get("port0", "mouse"))
    #     #self.options.append(self.port0_option)
    #
    #     self.amiga_config = config

    def prepare(self):
        print("AmigaRunner.prepare")

        # self.temp_dir = self.fsgs.temp_dir("amiga")

        # self.change_handler = GameChangeHandler(self.temp_dir)

        # self.firmware_dir = self.prepare_firmware("Amiga Firmware")
        # config = self.fsgs.config.copy()

        model = self.config["amiga_model"]
        if model.startswith("CD32"):
            platform = "CD32"
        elif model == "CDTV":
            platform = "CDTV"
        else:
            platform = "Amiga"
        # name = Settings.get("config_name")
        # name = self.fsgs.game.name

        # uuid = Config.get("x_game_uuid")
        # uuid = None

        from fsgs.SaveStateHandler import SaveStateHandler
        save_state_handler = SaveStateHandler(
            self.fsgs, self.get_name(), platform)

        self.config["joystick_port_0"] = self.ports[1].device_id or ""
        self.config["joystick_port_1"] = self.ports[0].device_id or ""

        self.launch_handler = LaunchHandler(
            self.fsgs, self.get_name(), self.config, save_state_handler)

        # self.change_handler.init(self.fsgs.get_game_state_dir(),
        #         ignore=["*.uss", "*.sdf"])

        # self.launch_handler.config["joystick_port_0"] = \
        #         self.inputs[1].device_id
        # self.launch_handler.config["joystick_port_1"] = \
        #         self.inputs[0].device_id

        if self.use_fullscreen():
            self.launch_handler.config["fullscreen"] = "1"
            if not self.launch_handler.config.get("fullscreen_mode", ""):
                if app.settings["fs-uae:fullscreen-mode::default"]:
                    self.launch_handler.config["fullscreen_mode"] = "window"
        else:
            self.launch_handler.config["fullscreen"] = "0"

        self.launch_handler.prepare()

    # def configure(self):
    #     print("AmigaGameHandler.configure")
    #
    #     #temp_dir = self.fsgs.temp.dir("amiga-config")
    #     #config_file = os.path.join(temp_dir, "amiga-config.fs-uae")
    #     #with open(config_file, "wb") as f:
    #     #    #self._configure_emulator(f)
    #     #    self.launch_handler.write_config(f)
    #     #    self.write_additional_config(f)#
    #     #
    #     ##self.args.extend(["--config", config_file])
    #     #self.args.extend([config_file])

    # def write_additional_config(self, f):
    #     #if self.get_option("fullscreen"):
    #     #    f.write("fullscreen = 1\n")
    #
    #     #if self.configure_vsync():
    #     #    f.write("video_sync = full\n")
    #     #else:
    #     #    f.write("video_sync = none\n")
    #
    #     if self.get_option("fsaa"):
    #         f.write("fsaa = {0}\n".format(str(self.get_option("fsaa"))))

    def run(self):
        print("AmigaGameHandler.run")
        # self.on_progress(_("Starting FS-UAE..."))
        config = self.launch_handler.create_config()
        process, self.temp_config_file = FSUAE.start_with_config(config)
        # process.wait()
        # print("LaunchHandler.start is done")
        # print("removing", config_file)
        # try:
        #     os.remove(config_file)
        # except Exception:
        #     print("could not remove config file", config_file)
        return process

    def finish(self):
        print("removing", self.temp_config_file)
        if self.temp_config_file is not None:
            try:
                os.remove(self.temp_config_file)
            except Exception:
                print("could not remove config file", self.temp_config_file)

        self.launch_handler.update_changes()
        self.launch_handler.cleanup()
        # self.change_handler.update(self.fsgs.get_game_state_dir())

    def get_supported_filters(self):
        supported = []
        # supported = [{
        #         "name": "2x",
        #         "gfx_filter": "none",
        #         "line_double": True,
        #         "lores": False,
        #     }, {
        #         "name": "none",
        #         "gfx_filter": "none",
        #         "line_double": False,
        #         "lores": False,
        #     }, {
        #         "name": "pal",
        #         "gfx_filter": "pal",
        #         "line_double": True,
        #         "lores": False,
        #     }, {
        #         "name": "scale2x",
        #         "gfx_filter": "scale2x",
        #         "line_double": False,
        #         "lores": True,
        #     }, {
        #         "name": "hq2x",
        #         "gfx_filter": "hq2x",
        #         "line_double": False,
        #         "lores": True,
        #     }
        # ]
        return supported

    def set_input_options(self, f):
        # FIXME

        input_mapping = [{
                "1": "JOY2_FIRE_BUTTON",
                "2": "JOY2_2ND_BUTTON",
                "3": "JOY2_3RD_BUTTON",
                "left": "JOY2_LEFT",
                "right": "JOY2_RIGHT",
                "up": "JOY2_UP",
                "down": "JOY2_DOWN",
            }, {
                "1": "JOY1_FIRE_BUTTON",
                "2": "JOY1_2ND_BUTTON",
                "3": "JOY1_3RD_BUTTON",
                "left": "JOY1_LEFT",
                "right": "JOY1_RIGHT",
                "up": "JOY1_UP",
                "down": "JOY1_DOWN",
            }]

        # f.write("input.config=1\n")

        f.write("\n[input]\n")
        # f.write("joyport0=none\n")
        # f.write("joyport1=none\n")
        # f.write("input.1.keyboard.0.empty=false\n")
        # f.write("input.1.keyboard.0.disabled=false\n")

        for i, input in enumerate(self.inputs):
            if not input.device:
                continue
            if i == 0:
                key = "joystick_port_1"
            else:
                key = "joystick_port_0"
            if input.type == "amiga_mouse":
                value = "mouse"
            else:
                value = input.device.id
            print("configure device:", key, value)
            f.write("{0} = {1}\n".format(key, value))

    def find_kickstart(self, bios_patterns):
        print("find_kickstart", bios_patterns)
        for name in os.listdir(self.firmware_dir):
            for bios in bios_patterns:
                if bios in name:
                    return os.path.join(self.firmware_dir, name)
        raise Exception("Could not find Amiga bios/kickstart " +
                repr(bios_patterns))

    def get_game_refresh_rate(self):
        # FIXME: Now assuming that all games are PAL / 50 Hz
        # - make configurable?
        return 50.0



import os
import time
from binascii import unhexlify
import sqlite3
# FIXME: remove dependency on launcher, have the Launcher tell this class
# the path instead
from fsgs.FSGSDirectories import FSGSDirectories
from .BaseDatabase import BaseDatabase


SENTINEL = "fae7671d-e232-4b71-b179-b3cd45995f92"
VERSION = 2


class File(dict):
    
    def __nonzero__(self):
        # noinspection PyTypeChecker
        return bool(self["path"])

    def __bool__(self):
        # noinspection PyTypeChecker
        return bool(self["path"])


class FileDatabase(BaseDatabase):

    def __init__(self, sentinel):
        BaseDatabase.__init__(self, sentinel)
        self.last_file_insert = None
        self.last_file_delete = None

    @classmethod
    def get_path(cls):
        # FIXME: remove direct dependency on launcher
        cache_dir = FSGSDirectories.get_cache_dir()
        path = os.path.join(cache_dir, "Files.sqlite")
        return path

    @classmethod
    def get_version(cls):
        return VERSION

    @classmethod
    def instance(cls, new=False):
        if new or not hasattr(cls.thread_local, "file_database"):
            cls.thread_local.file_database = cls(cls.SENTINEL)
        return cls.thread_local.file_database

    @classmethod
    def get_instance(cls):
        if not hasattr(cls.thread_local, "file_database"):
            cls.thread_local.file_database = cls(cls.SENTINEL)
        return cls.thread_local.file_database

    def get_file_ids(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT id FROM file WHERE parent IS NULL")
        return set([row[0] for row in cursor.fetchall()])

    def get_file_hierarchy_ids(self, path):
        path = self.encode_path(path)
        if path.endswith("/"):
            path = path[:-1]
        a_path = path + "\u002f"  # forward slash
        b_path = path + "\u0030"  # one more than forward slash
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT id FROM file WHERE path >= ? AND path < ?",
            (a_path, b_path))
        return set([row[0] for row in cursor.fetchall()])

    # def get_child_ids(self, id=id):
    #     cursor = self.internal_cursor()
    #     cursor.execute("SELECT id FROM file WHERE parent = ?", (id,))
    #     return set([row[0] for row in cursor.fetchall()])

    # def get_file_hierarchy(self, path):
    #     self.init()
    #     cursor = self.create_cursor()
    #     #query = "SELECT id, path FROM configuration WHERE path like ?"
    #     #args = ["$BASE/Configurations/%"]
    #     #self._cursor.execute(query, args)
    #     a_path = path + "\u002f" # forward slash
    #     b_path = path + "\u0030" # one more than forward slash
    #     query = "SELECT id, path FROM file WHERE " \
    #             "path >= ? AND path < ?"
    #     cursor.execute(query, (a_path, b_path))
    #     result = {}
    #     for row in cursor:
    #         result[decode_path(row[1])] = row[0]
    #     return result

    def encode_path(self, path):
        # this only works if both path and 
        # FSGSDirectories.base_dir (etc) have been
        # normalized with get_real_case.
        path = path.replace("\\", "/")
        base_dir = FSGSDirectories.get_base_dir()
        if path.startswith(base_dir):
            path = path[len(base_dir):]
            if path.startswith("/"):
                path = path[1:]
            path = "$/" + path
        return path
        # parts = path.split("/")
        # return "{0}{1}".format(len(parts), path)

    def decode_path(self, path):
        # path = path[1:]
        if not path or path[0] != "$":
            return path
        base_dir = FSGSDirectories.get_base_dir() + "/"
        # FIXME: $base_dir is deprecated
        # path = path.replace("$base_dir/", base_dir)
        # path = path.replace("$BASE/", base_dir)
        # path = path.replace("$/", base_dir)
        if path.startswith("$/"):
            path = base_dir + path[2:]
        return path

    def find_local_roms(self):
        if not self.connection:
            self.init()

        a = "$/Kickstarts/"
        b = "$/Kickstarts" + "\u0030"  # one more than forward slash
        query = "SELECT id, path FROM file WHERE path >= ? AND path < ?"
        # args = ["$BASE/Kickstarts/%"]
        cursor = self.internal_cursor()
        cursor.execute(query, (a, b))
        result = {}
        for row in cursor.fetchall():
            result[self.decode_path(row[1])] = row[0]
        return result

    def delete_file(self, id=None, path=None):
        cursor = self.internal_cursor()
        delete_ids = []
        if id is not None:
            delete_ids.append(id)
        if path is not None:
            path = self.encode_path(path)
            cursor.execute("SELECT id FROM file WHERE path = ?", (path,))
            for row in cursor:
                delete_ids.append(row[0])            
        for id in delete_ids:
            cursor.execute(
                "DELETE FROM file WHERE id = ? OR parent = ?", (id, id))
        self.last_file_delete = int(time.time())

    def check_sha1(self, sha1):
        # if not hasattr(self, "check_sha1_cursor"):
        #     #self.check_sha1_cursor = cursor()
        #
        #     self.init()
        #     self.check_sha1_cursor = self.create_cursor()
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT count(*) FROM file WHERE sha1 = ?", (
                sqlite3.Binary(unhexlify(sha1)),))
        return cursor.fetchone()[0]

    def find_file(self, name="", sha1="", path=""):
        cursor = self.internal_cursor()
        if sha1:
            sha1 = unhexlify(sha1.encode("ASCII"))
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE sha1 = ? LIMIT 1",
                (sqlite3.Binary(sha1),))
        elif name:
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE name = ? COLLATE NOCASE LIMIT 1",
                (name.lower(),))
        else:
            path = self.encode_path(path)
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE path = ? LIMIT 1", (path,))
        row = cursor.fetchone()
        result = File()
        if row:
            if row[5]:
                # parent
                cursor.execute(
                    "SELECT path FROM file WHERE id = ?", (row[5],))
                path = cursor.fetchone()[0] + row[1]
            else:
                path = row[1]
            path = self.decode_path(path)
            # if result is not None:
            result["id"] = row[0]
            result["path"] = path
            result["sha1"] = row[2]
            result["mtime"] = row[3]
            result["size"] = row[4]
            # return result
        else:
            # if result is not None:
            result["id"] = None
            result["path"] = None
            result["sha1"] = None
            result["mtime"] = None
            result["size"] = None
            # return None
        return result

    def find_files(self, ext=None):
        cursor = self.internal_cursor()
        query = "SELECT path FROM file WHERE 1 = 1"
        args = []
        if ext is not None:
            # this is used (for now) to look up .fs-uae files, so..
            # we don't want files from archives
            # sqlite like is case insensitive by default
            query += " AND parent is NULL AND path LIKE ?"
            args.append("%" + ext)
        cursor.execute(query, args)
        results = []
        for row in cursor:
            data = {
                "path": self.decode_path(row[0]),
                # "name": row[1]
            }
            results.append(data)
        return results

    def add_file(self, path="", sha1=None, mtime=0, size=0, parent=None):
        self.init()
        # if not name:
        #     name = os.path.basename(path)
        path = self.encode_path(path)
        # parent_path, name = path.rsplit("/", 1)

        cursor = self.internal_cursor()
        # parent = self.ensure_dir(dir)
        print("add_file", repr(path), repr(sha1), repr(mtime), repr(size),
              repr(parent))
        if sha1:
            sha1 = unhexlify(sha1)
        cursor.execute(
            "INSERT INTO file (path, sha1, mtime, size, parent) "
            "VALUES (?, ?, ?, ?, ?)",
            (path, sqlite3.Binary(sha1), mtime, size, parent))
        self.last_file_insert = int(time.time())
        return cursor.lastrowid

    # def ensure_dir(self, path):
    #     #if not path.endswith("/"):
    #     #    path += "/"
    #     assert path.endswith("/")
    #     cursor = self.internal_cursor()
    #     cursor.execute("SELECT id FROM file WHERE name = ?", (path,))
    #     row = cursor.fetchone()
    #     if row is not None:
    #         return row[0]
    #     cursor.execute("INSERT INTO file (name) VALUES (?)", (path,))
    #     return cursor.lastrowid

    def get_last_event_stamps(self):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT last_file_insert, last_file_delete FROM metadata")
        row = cursor.fetchone()
        result = {
            "last_file_insert": row[0],
            "last_file_delete": row[0],
        }
        return result

    def update_last_event_stamps(self):
        if self.last_file_insert:
            cursor = self.internal_cursor()
            cursor.execute(
                "UPDATE metadata set last_file_insert = ?",
                (self.last_file_insert,))
            self.last_file_insert = None
        if self.last_file_delete:
            cursor = self.internal_cursor()
            cursor.execute(
                "UPDATE metadata set last_file_delete = ?",
                (self.last_file_delete,))
            self.last_file_delete = None

    def clear(self):
        if not self.connection:
            self.init()
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM file")

    def update_database_to_version_1(self):
        cursor = self.internal_cursor()
        try:
            cursor.execute("SELECT count(*) FROM file")
        except sqlite3.OperationalError:
            cursor.execute("""CREATE TABLE file (
                    id INTEGER PRIMARY KEY,
                    sha1 BLOB,
                    path TEXT,
                    size INTEGER,
                    mtime INTEGER,
                    parent INTEGER
                    )""")
            cursor.execute("CREATE INDEX file_sha1 ON file(sha1)")
            cursor.execute("CREATE INDEX file_path ON file(path)")

        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_insert INTEGER")
        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_delete INTEGER")

    def update_database_to_version_2(self):
        cursor = self.internal_cursor()
        cursor.execute("CREATE INDEX file_parent ON file(parent)")

    # def update_database_to_version_1(self):
    #     cursor = self.create_cursor()
    #     try:
    #         cursor.execute("SELECT count(*) FROM file")
    #     except sqlite3.OperationalError:
    #         cursor.execute("""CREATE TABLE file (
    # id integer primary key,
    # path text,
    # name text,
    # sha1 text,
    # size integer,
    # mtime integer
    # )""")
    #         cursor.execute("CREATE INDEX file_sha1 ON file(sha1)")
    #         cursor.execute("CREATE INDEX file_path ON file(path)")
    #         cursor.execute("CREATE INDEX file_name ON file("
    #                        "name collate nocase)")
    #
    # def update_database_to_version_2(self):
    #     cursor = self.create_cursor()
    #     cursor.execute("SELECT id, sha1 FROM file WHERE sha1 IS NOT NULL")
    #     for row in cursor.fetchall():
    #         id, sha1 = row
    #         sha1 = unhexlify(sha1)
    #         cursor.execute("UPDATE file SET sha1 = ? WHERE id = ?",
    #                 (sqlite3.Binary(sha1), id))
    #
    # def update_database_to_version_3(self):
    #     cursor = self.create_cursor()
    #     cursor.execute("ALTER TABLE metadata ADD COLUMN last_file_insert INT")
    #     cursor.execute("ALTER TABLE metadata ADD COLUMN last_file_delete INT")



import fsui as fsui
from ...DeviceManager import DeviceManager
from ...I18N import gettext
from ...Settings import Settings


joystick_mode_values = ["nothing", "mouse", "joystick"]
joystick_values = ["none", "mouse", "keyboard"]


# def get_joystick_mode_titles():
#     return [gettext("Nothing"), gettext("Mouse"), gettext("Joystick")]


class PreferredJoysticksGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()
        # self.layout.padding_left = 10
        # self.layout.padding_right = 10

        # image = fsui.Image("fs_uae_launcher:res/joystick.png")
        # self.image_view = fsui.ImageView(self, image)
        # self.layout.add(self.image_view, valign=0.0)

        # self.layout.add_spacer(20)

        self.layout2 = fsui.VerticalLayout()
        self.layout.add(self.layout2, fill=True, expand=True)

        heading = gettext("Preferred Joysticks")
        label = fsui.HeadingLabel(self, heading)
        self.layout2.add(label)

        self.layout2.add_spacer(20)
        label = fsui.Label(self, gettext("The following joystick will be "
                                         "preferred, if present:"))
        self.layout2.add(label)

        self.layout2.add_spacer(6)
        selector = PreferredJoystickSelector(self, 0)
        self.layout2.add(selector, fill=True)

        self.layout2.add_spacer(20)
        label = fsui.Label(
            self, gettext("Preferred device for secondary joystick:"))
        self.layout2.add(label)

        self.layout2.add_spacer(6)
        selector = PreferredJoystickSelector(self, 1)
        self.layout2.add(selector, fill=True)


class PreferredJoystickSelector(fsui.Group):

    def __init__(self, parent, index):
        self.index = index
        if index:
            self.key = "secondary_joystick"
        else:
            self.key = "primary_joystick"

        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        devices = ["", get_keyboard_title()]
        # for i, name in enumerate(DeviceManager.get_joystick_names()):
        #     devices.append(name)
        for device_name in DeviceManager.get_joystick_names():
            if DeviceManager.is_joystick(device_name):
                devices.append(device_name)

        self.device_choice = fsui.ComboBox(self, devices)

        self.layout.add(self.device_choice, expand=True)

        # Config.add_listener(self)

        self.initialize_from_settings()
        self.set_settings_handlers()

    def initialize_from_settings(self):
        self.on_setting(self.key, Settings.get(self.key))

    def set_settings_handlers(self):
        self.device_choice.on_change = self.on_device_change
        Settings.add_listener(self)

    def on_destroy(self):
        Settings.remove_listener(self)

    def on_device_change(self):
        value = self.device_choice.get_text()
        print("on_device_change", value)
        if value == get_keyboard_title():
            value = "keyboard"
        Settings.set(self.key, value)

    def on_setting(self, key, value):
        if key == self.key:
            if value == "keyboard":
                value = get_keyboard_title()
            self.device_choice.set_text(value)


def get_keyboard_title():
    return gettext("Cursor Keys and Right Ctrl/Alt")



import os
from uuid import uuid4, uuid5, NAMESPACE_URL
import shutil
from urllib.request import urlopen
import hashlib
# FIXME: temporary package dependency cycle, must be fixed
from fsgs.FSGSDirectories import FSGSDirectories


class Downloader(object):

    @classmethod
    def check_terms_accepted(cls, download_file, download_terms):
        print("check_terms_accepted", download_file, download_terms)
        uuid = str(uuid5(NAMESPACE_URL, download_file + download_terms))
        path = cls.get_cache_path(uuid)
        print("check_terms_accepted", path)
        return os.path.exists(path)

    @classmethod
    def set_terms_accepted(cls, download_file, download_terms):
        print("set_terms_accepted", repr(download_file), repr(download_terms))
        uuid = str(uuid5(NAMESPACE_URL, download_file + download_terms))
        path = cls.get_cache_path(uuid)
        print("set_terms_accepted", path)
        with open(path, "wb") as _:
            pass

    @classmethod
    def cache_data(cls, data):
        sha1 = hashlib.sha1(data).hexdigest()
        cache_path = cls.get_cache_path(sha1)
        if os.path.exists(cache_path):
            return
        cache_path_temp = cache_path + ".partial." + str(uuid4())
        with open(cache_path_temp, "wb") as f:
            f.write(data)
        try:
            os.rename(cache_path_temp, cache_path)
        except Exception as e:
            if os.path.exists(cache_path):
                # file has appeared in the meantime (by another thread?)
                return
            raise e

    @classmethod
    def cache_file_from_url(cls, url, download=True, opener=None):
        print("DownloadService.cache_file_from_url", url)

        cache_path = cls.get_url_cache_path(url)
        if os.path.exists(cache_path):
            print("(in cache)")
            # so we later can delete least accessed files in cache...
            os.utime(cache_path, None)
            return cache_path
        if not download:
            return None
        if opener:
            ifs = opener.open(url)
        else:
            ifs = urlopen(url)
        cache_path_temp = cache_path + ".partial." + str(uuid4())
        with open(cache_path_temp, "wb") as ofs:
            while True:
                data = ifs.read(65536)
                if not data:
                    break
                ofs.write(data)
        os.rename(cache_path_temp, cache_path)
        return cache_path

    @classmethod
    def install_file_from_url(cls, url, path):
        print("DownloadService.install_file_from_url", url)
        print(repr(path))
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        cache_path = cls.cache_file_from_url(url)
        temp_path = path + ".partial"
        shutil.copy(cache_path, temp_path)
        os.rename(temp_path, path)

    @classmethod
    def install_file_by_sha1(cls, sha1, name, path):
        print("DownloadService.install_file_by_sha1", sha1)
        print(repr(path))
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        cache_path = cls.get_cache_path(sha1)
        if os.path.exists(cache_path):
            print("(in cache)")
            shutil.copy(cache_path, path)
            # so we later can delete least accessed files in cache...
            os.utime(cache_path, None)
            return
        url = cls.sha1_to_url(sha1, name)
        ifile = urlopen(url)
        temp_path = path + ".partial." + str(uuid4())
        h = hashlib.sha1()
        with open(temp_path, "wb") as ofile:
            while True:
                data = ifile.read(65536)
                if not data:
                    break
                h.update(data)
                ofile.write(data)
        if h.hexdigest() != sha1:
            print("error: downloaded sha1 is", sha1)
            raise Exception("sha1 of downloaded file does not match")
        temp_cache_path = cache_path + ".partial." + str(uuid4())
        shutil.copy(temp_path, temp_cache_path)
        os.rename(temp_cache_path, cache_path)
        os.rename(temp_path, path)

    @classmethod
    def sha1_to_url(cls, sha1, name):
        url = "http://fs-uae.net/s/sha1/{0}/{1}".format(sha1, name)
        print(url)
        return url

    @classmethod
    def get_cache_path(cls, sha1_or_uuid):
        path = os.path.join(
            FSGSDirectories.get_cache_dir(), "Downloads", sha1_or_uuid[:3])
        if not os.path.exists(path):
            os.makedirs(path)
        return os.path.join(path, sha1_or_uuid)

    @classmethod
    def get_url_cache_path(cls, url):
        if isinstance(url, str):
            # Python 3
            return cls.get_cache_path(str(uuid5(NAMESPACE_URL, url)))
        else:
            # Python 2 / unicode
            return cls.get_cache_path(
                str(uuid5(NAMESPACE_URL, url.encode("UTF-8"))))



from fsbc.util import unused
from fsui import Panel, Color, VerticalLayout, HorizontalLayout, Image
from ...Config import Config
from ..Skin import Skin
from .LanguageElement import LanguageElement
from .PlayersElement import PlayersElement
from .ProtectionElement import ProtectionElement
from .WarningsElement import WarningsElement
from .WebLinkElement import WebLinkElement


class StatusBar(Panel):

    def __init__(self, parent):
        Panel.__init__(self, parent, paintable=True)
        self.set_min_height(29)
        # self.set_background_color(Color(0xd8, 0xd8, 0xd8))
        self.layout = VerticalLayout()
        self.hori_layout = HorizontalLayout()
        self.layout.add(self.hori_layout, fill=True, expand=True,
                        margin_top=1)

        element = ProtectionElement(self)
        self.hori_layout.add(element, fill=True)

        element = WarningsElement(self)
        self.hori_layout.add(element, fill=True, expand=True)
        self.hori_layout.add_spacer(16)

        for language, icon_name in reversed([
            ("en", "flag-gb"),
            ("de", "flag-de"),
            ("fr", "flag-fr"),
            ("es", "flag-es"),
            ("it", "flag-it"),
            ("ja", "flag-jp"),
            # ("", "flag-unknown"),
        ]):
            icon = Image("fs_uae_workspace:res/16/" + icon_name + ".png")
            element = LanguageElement(self, language, icon)
            self.hori_layout.add(element, fill=True)
        self.hori_layout.add_spacer(16)

        element = PlayersElement(self)
        self.hori_layout.add(element, fill=True)

        # for config_key, icon_name in [
        #     ("database_url", "database_url_16"),
        #     ("hol_url", "hol_url_16"),
        #     ("lemonamiga_url", "lemon_url_16"),
        #     ("mobygames_url", "mobygames_url_16"),
        #     ("wikipedia_url", "wikipedia_url_16"),
        # ]:
        #     icon = Image("fs_uae_launcher:res/" + icon_name + ".png")
        #     element = LinkButtonElement(self, config_key, icon)
        #     self.hori_layout.add(element)

        element = WebLinkElement(self)
        self.hori_layout.add(element, fill=True)

        # this listener is added after all status bar children have
        # added their listeners, this is important for re-layout...
        Config.add_listener(self)

    def on_destroy(self):
        Config.remove_listener(self)

    def on_config(self, key, value):
        unused(value)
        layout = False
        if key in ["languages", "protection"]:
            layout = True
        if layout:
            self.layout.update()

    def on_paint(self):
        dc = self.create_dc()
        color_1 = Skin.get_background_color()
        if color_1 is not None:
            color_1 = color_1.copy().darken(0.12)
        else:
            color_1 = Color(0x00, 0x00, 0x00, 0x30)
        dc.draw_line(0, 0, self.size[0], 0, color_1)
        self.draw_background(self, dc, offset=1, height=self.size[1] - 1)

    @classmethod
    def draw_element_background(cls, widget, dc):
        cls.draw_background(widget, dc)

    @classmethod
    def draw_background(cls, widget, dc, offset=None, height=None):
        x = 0
        y = 0
        w = widget.size[0]
        h = widget.size[1]
        if offset is not None:
            y += offset
        if height is not None:
            h = height

        color_1 = Skin.get_background_color()
        color_2 = color_1
        # if fsui.System.macosx:
        #     color_1 = Color(0xa7, 0xa7, 0xa7)
        #     color_2 = Color(0xc0, 0xc0, 0xc0)
        if color_1 is not None:
            color_1 = color_1.copy().darken(0.08)
        else:
            color_1 = Color(0x00, 0x00, 0x00, 0x20)
            color_2 = Color(0x00, 0x00, 0x00, 0x00)
        dc.draw_vertical_gradient(x, y, w, h, color_1, color_2)



from fsgs.mednafen.mednafen import MednafenRunner


class GameGearRunner(MednafenRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Built-in Controller",
        "mapping_name": "gamegear",
    }

    PORTS = [
        {
            "description": "Controller",
            "types": [CONTROLLER]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def mednafen_system_prefix(self):
        return "gg"

    def mednafen_input_mapping(self, _):
        return {
            "1": "gg.input.builtin.gamepad.button1",
            "2": "gg.input.builtin.gamepad.button2",
            "UP": "gg.input.builtin.gamepad.up",
            "DOWN": "gg.input.builtin.gamepad.down",
            "LEFT": "gg.input.builtin.gamepad.left",
            "RIGHT": "gg.input.builtin.gamepad.right",
            "START": "gg.input.builtin.gamepad.start",
        }



import os
import sys
import time
import fsui
import weakref
from fsui.qt import Qt, QCursor
import traceback
import threading
from fsbc.system import macosx, windows
from collections import deque
from game_center.Application import Application
from fsbc.Application import app
from game_center.glui.imageloader import ImageLoader
from game_center.glui.input import InputHandler
from .gnome3 import running_in_gnome_3, handle_gnome_extensions


class MainQueueItem(object):
    done = False
    exception = None
    result = None


_main_thread = threading.currentThread()


class Main(object):

    queue = deque()
    lock = threading.Lock()

    @classmethod
    def process(cls):
        with cls.lock:
            for i in range(len(cls.queue)):
                item = cls.queue.popleft()
                try:
                    item.result = item.func(*item.args, **item.kwargs)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    item.exception = e
                item.done = True

    @classmethod
    def call(cls, func, *args, **kwargs):
        if threading.currentThread() == _main_thread:
            return func(*args, **kwargs)
        item = MainQueueItem()
        item.func = func
        item.args = args
        item.kwargs = kwargs
        with cls.lock:
            cls.queue.append(item)
        while not item.done:
            time.sleep(0.01)
        if item.exception:
            raise item.exception
        return item.result


class Callbacks():

    def __init__(self):
        self.width = 320
        self.height = 200
        self.window = None

    def initialize(self):
        from game_center.gamecenter import GameCenter
        GameCenter.init()

        import game_center.glui
        game_center.glui.main(g_callbacks)

    def resize(self, width, height):
        print("Callbacks.resize", width, height)
        from game_center.glui.window import on_resize
        self.width = width
        self.height = height
        on_resize((width, height))

    def render(self):
        from game_center.glui.window import main_loop_iteration
        if main_loop_iteration():
            main_window.close()

    def timer(self):
        InputHandler.update()

    def restore_window_if_necessary(self):
        pass
        # noinspection PyCallingNonCallable
        # window = self.window()
        # under Gnome 3, the window is minized when launching FS-UAE
        # full-screen from full-screen arcade/game center.
        # window.restore_window_if_necessary()

    def set_window(self, window):
        self.window = weakref.ref(window)


def check_setting(name):
    name = name.replace("_", "-")
    if "--" + name in sys.argv:
        return "1"
    if "--" + name + "=1" in sys.argv:
        return "1"
    if "--no-" + name in sys.argv:
        return "0"
    if "--" + name + "=0" in sys.argv:
        return "0"
    return ""


def main():
    global main_window, g_callbacks
    # import logging; logging.shutdown(); import sys; sys.exit()
    print("Fengestad Game System...")
    application = Application()
    fs_width, fs_height = fsui.get_screen_size()

    cursor_position = None
    use_window = False
    use_window_decorations = True
    use_fullscreen = True
    use_fullscreen_window = False
    use_top_clock = check_setting("top_clock") != "0"
    use_top_logo = check_setting("top_logo") != "0"

    if running_in_gnome_3():
        if check_setting("fullscreen") == "0":
            # we want a normal window
            pass
        else:
            use_fullscreen = False
            use_window_decorations = False
            use_window = "maximized"
            if check_setting("gnome_extensions") != "0":
                handle_gnome_extensions()
            cursor_position = fs_width - 1, fs_height - 1
        # use_top_clock = False
        # use_top_logo = False
        # app.settings["fs-uae:fullscreen-mode::default"] = "window"

    if windows:
        pass
    elif macosx:
        use_fullscreen_window = True
        app.settings["fs-uae:fullscreen-mode::default"] = "window"
    else:
        # app.settings["fs-uae:fullscreen-mode::default"] = "window"
        pass

    if check_setting("fullscreen"):
        use_fullscreen = check_setting("fullscreen") == "1"

    if "--fullscreen-mode=fullscreen" in sys.argv:
        use_fullscreen_window = False
    elif "--fullscreen-mode=window" in sys.argv:
        use_fullscreen_window = True

    if "--maximize" in sys.argv:
        use_window = "maximized"
        use_fullscreen = False

    if "--no-window-decorations" in sys.argv:
        use_window_decorations = False

    app.settings["game-center:fullscreen"] = \
        "1" if use_fullscreen else "0"
    if use_fullscreen_window:
        app.settings["game-center:fullscreen-mode"] = "window"
    else:
        app.settings["game-center:fullscreen-mode"] = ""
    app.settings["game-center:window-decorations"] = \
        "1" if use_window_decorations else "0"
    app.settings["game-center:maximize"] = \
        "1" if use_window == "maximized" else "0"
    app.settings["game-center:top-clock"] = "1" if use_top_clock else "0"
    app.settings["game-center:top-logo"] = "1" if use_top_logo else "0"

    monitor = "middle-left"
    for arg in sys.argv:
        if arg.startswith("--monitor="):
            monitor = arg[10:]
            app.settings["monitor"] = monitor
            break
    else:
        if app.settings["monitor"]:
            monitor = app.settings["monitor"]

    if macosx and use_fullscreen and use_fullscreen_window:
        # noinspection PyUnresolvedReferences
        import objc
        # noinspection PyUnresolvedReferences
        from Foundation import NSBundle
        bundle = NSBundle.bundleWithPath_(
            "/System/Library/Frameworks/Carbon.framework")
        objc.loadBundleFunctions(
            bundle, globals(), (("SetSystemUIMode", b"III", ""),))
        # kUIModeAllHidden = 3
        # kUIOptionAutoShowMenuBar = 1 << 0
        # noinspection PyUnresolvedReferences
        SetSystemUIMode(3, 1 << 0)

    from game_center.qtui.qtwindow import QtWindow
    # main_window = QtWindow(timer_callback, 16)
    g_callbacks = Callbacks()
    main_window = QtWindow(g_callbacks, 16)
    g_callbacks.set_window(main_window)

    if use_fullscreen:
        # must move the cursor after the window is created
        if cursor_position is not None:
            QCursor.setPos(cursor_position[0], cursor_position[1])
            # os.environ["FSGS_RETURN_CURSOR_TO"] = "{0},{1}".format(
            #     cursor_position[0], cursor_position[1])

    # main_window.show()
    if use_fullscreen:
        # fs_width += 1
        main_window.resize(fs_width, fs_height)

        desktop = application.qapplication.desktop()
        screens = []
        for i in range(desktop.screenCount()):
            geometry = desktop.screenGeometry(i)
            screens.append([geometry.x(), i, geometry])
        screens.sort()
        if monitor == "left":
            mon = 0
        elif monitor == "middle-right":
            mon = 2
        elif monitor == "right":
            mon = 3
        else:  # middle-left
            mon = 1
        display = round(mon / 3 * (len(screens) - 1))
        geometry = screens[display][2]

        main_window.setGeometry(geometry)

        if use_fullscreen_window:
            print("using fullscreen window")

            # main_window.setWindowFlags(Qt.X11BypassWindowManagerHint)

            # if "--always-on-top" in sys.argv:
            #     print("set always on top")
            #     main_window.setWindowFlags(
            #         main_window.windowFlags() | Qt.WindowStaysOnTopHint)

            main_window.setWindowFlags(
                main_window.windowFlags() | Qt.FramelessWindowHint)
            main_window.show()
            # main_window.activateWindow()
        else:
            main_window.showFullScreen()
        # main_window.windowHandle().setScreen(screen)
    else:
        if not use_window_decorations:
            main_window.setWindowFlags(Qt.FramelessWindowHint)

        if use_window == "maximized":
            main_window.showMaximized()
        else:
            main_window.resize(960, 540)
            main_window.show()

    if cursor_position is not None:
        os.environ["FSGS_RETURN_CURSOR_TO"] = "{0},{1}".format(
            cursor_position[0], cursor_position[1])

    # main_window.setTim
    # main_loop_iteration

    application.run()
    print("application.run returned")

    application.stop()
    ImageLoader.get().stop()

    application.wait()

    print(" --- game_center.glui.main is done ---")
    print(application.name)
    return



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013  Frode Solheim <frode-code@fengestad.no>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Game Runner for Commodore 64.
"""

import os
import io
from fsgs.runner import GameRunner
from fsgs.input.mapper import InputMapper


VIDEO_WIDTH = 384
VIDEO_HEIGHT = 272


# noinspection PyAttributeOutsideInit
class Commodore64Runner(GameRunner):

    def prepare(self):
        self.temp_dir = self.create_temp_dir("vice")
        config_file = os.path.join(self.temp_dir.path, "vice.cfg")
        with io.open(config_file, "w", encoding="UTF-8") as f:
            self.vice_configure(f)
            # self.configure_input(f)
        self.add_arg("-config", config_file)

    def run(self):
        # return self.start_emulator("fs-vice/x64sc")
        return self.start_emulator_from_plugin_resource("x64sc")

    def finish(self):
        pass

    def vice_configure(self, f):
        f.write("[C64SC]\n")
        f.write("ConfirmOnExit=0\n")
        # f.write("KeepAspectRatio=1\n")
        f.write("VICIIHwScale=1\n")
        f.write("HwScalePossible=1\n")
        f.write("VICIIDoubleScan=1\n")
        f.write("VICIISDLFullscreenMode=1\n")
        f.write("AutostartWarp=0\n")
        f.write("Drive8Type=1541\n")
        f.write("VICIIFilter=0\n")
        f.write("\n")

        screen_w, screen_h = self.screen_size()
        f.write("SDLCustomWidth={w}\n".format(w=screen_w))
        f.write("SDLCustomHeight={h}\n".format(h=screen_h))

        original_aspect = VIDEO_WIDTH / VIDEO_HEIGHT
        # original_aspect = 4 / 3
        # original_aspect = 320 / 272
        # original_aspect = 384 / 272
        # original_aspect = 1.0

        if self.use_fullscreen():
            f.write("VICIIFullscreen=1\n")

            if self.use_stretching():
                screen_aspect = screen_w / screen_h
                custom_aspect = screen_aspect / original_aspect
                print("custom aspect is", custom_aspect)
                f.write("AspectRatio={0}\n".format(custom_aspect))
                f.write("SDLGLAspectMode=1\n")

            # f.write("FullscreenEnabled=1\n")
            # if fs.linux:
            #     self.args.append("-VICIIhwscale")
            #     hwscoords = self.set_hwscale_coords()
            # else:
            #     pass
            #     #if self.options["zoom"]:
            #     #    logger.warning("zoom is not supported on this platform")
            #     #if self.options["stretch"]:
            #     #    logger.warning("stretch is not supported on this platform")
            #     #if self.options["border"]:
            #     #    logger.warning("border is not supported on this platform")
            # if fs.macosx:
            #     #logger.warning("Fullscreen not supported on Mac OS X")
            #     pass
            # else:
            # self.add_arg("-fullscreen")

        if self.use_audio_frequency():
            f.write("SoundSampleRate={0}\n".format(self.use_audio_frequency()))
        # default buffer size for vice is 100ms, that's far too much...
        # f.write("SoundBufferSize={0}\n".format(20))

        # print("set_media_options")
        # media_list = self.create_media_list()
        # print("sort media list")
        # # FIXME: SORT ON name (lowercase) ONLY, not whole path, because
        # # some files may have moved to another dir (temp dir)
        # media_list = sorted(media_list)
        # print(media_list)
        # if media_list[0].lower()[-4:] == ".crt":
        #     f.write("CartridgeFile=\"{path}\"\n".format(path=media_list[0]))
        #     f.write("CartridgeType={type}\n".format(type=0))
        #     f.write("CartridgeMode={mode}\n".format(mode=0))
        #     f.write("CartridgeReset={reset}\n".format(reset=1))
        # #    #self.args.extend(["-autostart", media_list[0]])
        # else:
        #     #f.write('AutostartPrgMode=2\n') # disk image
        #     #f.write('AutostartPrgDiskImage="{path}"\n'.format(
        #     #        path=media_list[0]))
        #     self.add_arg("-autostart", media_list[0])

        self.media_dir = self.create_temp_dir("vice-media")

        file_uri = self.config["tape_drive"]
        input_stream = self.fsgs.file.open(file_uri)
        tape_file = os.path.join(self.media_dir.path, file_uri.split("/")[-1])
        with open(tape_file, "wb") as tape_f:
            tape_f.write(input_stream.read())
        self.add_arg("-autostart", tape_file)

        # # FIXME: Floppies?
        # print(media_list)
        # if media_list[0].lower()[-4:] == ".d64":
        #     f = self.context.temp.file('fliplist')
        #     f.write("# Vice fliplist file\n\n")
        #     f.write("UNIT 8\n")
        #     print("FLIP LIST:")
        #     # Make sure first disk is added to the end of the fliplist
        #     for floppy in (media_list[1:] + [media_list[:1]]):
        #         print("%s\n" % (floppy,))
        #         print
        #         f.write("%s\n" % (floppy,))
        #     f.close()
        #     self.add_arg("-flipname", self.context.temp.file("fliplist"))
        #
        # media_dir = os.path.dirname(media_list[0])
        # print(media_dir)
        # f.write("InitialDefaultDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialTapeDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialCartDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialDiskDir=\"{dir}\"\n".format(dir=media_dir))
        # f.write("InitialAutostartDir=\"{dir}\"".format(dir=media_dir))

    # def vice_prepare_floppies(self):
    #     floppies = []
    #     #media_dir = os.path.dirname(self.context.game.file)
    #     #base_match = self.extract_match_name(os.path.basename(
    #     #        self.context.game.file))
    #     #for name in os.listdir(media_dir):
    #     #    dummy, ext = os.path.splitext(name)
    #     #    if ext.lower() not in ['.st']:
    #     #        continue
    #     #    match = self.extract_match_name(name)
    #     #    if base_match == match:
    #     #        floppies.append(os.path.join(media_dir, name))
    #     #        #floppies.append(name)
    #     if self.config["floppy_drive_0"]:
    #         floppies.append(self.config["floppy_drive_0"])
    #     return floppies


class ViceInputMapper(InputMapper):

    def axis(self, axis, positive):
        axis_start = self.device.buttons
        offset = 1 if positive else 0
        return "input.1.joystick.{0}.button.{1}".format(
            self.device.index, axis_start + axis * 2 + offset)

    def hat(self, hat, direction):
        hat_start = self.device.buttons + self.device.axes * 2
        offset = {
            "left": 0,
            "right": 1,
            "up": 2,
            "down": 3,
        }[direction]
        return "input.1.joystick.{0}.button.{1}".format(
            self.device.index, hat_start + hat * 4 + offset)

    def button(self, button):
        return "input.1.joystick.{0}.button.{1}".format(
            self.device.index, button)

    def key(self, key):
        return "input.1.keyboard.0.button.{0}.{1}".format(
            key.dinput_code, key.dinput_name[4:])



from fsui.qt import QWidget, QVBoxLayout
from .desktopwidget import DesktopWidget
from .taskbarwidget import TaskBarWidget
from .titlebarwidget import TitleBarWidget


class ShellWidget(QWidget):

    def __init__(self, parent, window_handler):
        super().__init__(parent)

        self.title_bar = TitleBarWidget(self, window_handler)
        self.desktop = DesktopWidget(self)
        self.task_bar = TaskBarWidget(self, window_handler)

        self.layout = QVBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.title_bar)
        self.layout.addWidget(self.desktop, 1)
        self.layout.addWidget(self.task_bar)



import time
from .base import SynchronizerBase
from ..LockerDatabase import LockerDatabase
from fsgs.FileDatabase import FileDatabase
from fsgs.res import gettext
import fsbc.Settings


def is_locker_enabled():
    return fsbc.Settings.get("database_locker") != "0"


class LockerSynchronizer(SynchronizerBase):

    def __init__(self, *args, **kwargs):
        SynchronizerBase.__init__(self, *args, **kwargs)

    def synchronize(self):
        if not is_locker_enabled():
            return

        if "locker-sync" not in self.context.meta:
            # we haven't looked up synchronization information from the server,
            # that probably means we didn't want to synchronize with the
            # server now, therefore we just return
            return

        if self.stop_check():
            return

        database = LockerDatabase.instance()
        sync_version = database.get_sync_version()
        if sync_version == self.context.meta["locker-sync"]:
            print("locker already up to date")
            return

        self.set_status(gettext("Fetching locker data..."))
        data = self.fetch_data("/api/locker-sync/1")
        assert len(data) % 20 == 0

        self.set_status(gettext("Updating locker data..."))
        database.clear()

        k = 0
        while k < len(data):
            sha1_bytes = data[k:k + 20]
            database.add_sha1_binary(sha1_bytes)
            k += 20

        database.set_sync_version(self.context.meta["locker-sync"])

        self.set_status(gettext("Committing locker data..."))

        # This isn't very elegant, but in order to force the game scanner to
        # refresh the game list based on files, we update the stamps in the
        # file database. Also, since we haven't keep track of additions /
        # deletions, we set both stamps for now...

        file_database = FileDatabase.instance()
        file_database.last_file_insert = time.time()
        file_database.last_file_delete = time.time()
        file_database.update_last_event_stamps()
        file_database.commit()

        database.commit()



import traceback
from fsbc.util import unused
from fsui.qt import Qt, QWidget, QPainter
from .Widget import Widget


class Panel(QWidget, Widget):

    def __init__(self, parent, paintable=False):
        unused(paintable)
        QWidget.__init__(self, parent.get_container() if parent else None)
        self.init_widget(parent)
        self.layout = None
        self._painter = None

        # QWidget.__init__(self)
        # self.setParent(parent.get_container())
        # super(Panel, self).__init__(parent.get_container())
        # super().__init__()
        self.move(0, 2000)
        self.setAutoFillBackground(True)

    @property
    def size(self):
        return self.get_size()

    def on_left_down(self):
        pass

    def on_left_up(self):
        pass

    def on_left_dclick(self):
        pass

    def on_paint(self):
        pass

    def create_dc(self):
        from .DrawingContext import DrawingContext
        return DrawingContext(self._painter)

    def paintEvent(self, event):
        # if not self._paintable:
        #     #dc = wx.PaintDC(self)
        #     return
        
        self._painter = QPainter(self)
        self._painter.setRenderHint(QPainter.Antialiasing)
        try:
            self.on_paint()
        except Exception:
            traceback.print_exc()
        finally:
            self._painter = None

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.on_left_down()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.on_left_up()

    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.on_left_dclick()

    def showEvent(self, event):
        self.on_resize()

    def resizeEvent(self, event):
        self.on_resize()

    def enterEvent(self, _):
        self.on_mouse_enter()

    def leaveEvent(self, _):
        self.on_mouse_leave()

    def on_mouse_enter(self):
        pass

    def on_mouse_leave(self):
        pass



import os
import hashlib
import traceback
from fsgs.FileDatabase import FileDatabase
from fsgs.Archive import Archive
from fsgs.amiga.ROMManager import ROMManager
from fsgs.ogd.client import OGDClient
from .I18N import gettext
from .Util import get_real_case


class FileScanner(object):

    def __init__(self, paths, purge_other_dirs,
                 on_status=None, on_rom_found=None, stop_check=None):
        self.paths = paths

        self.purge_other_dirs = purge_other_dirs
        self.database_file_ids = []

        self.on_status = on_status
        self.on_rom_found = on_rom_found
        self._stop_check = stop_check
        self.scan_count = 0
        self.files_added = 0
        self.bytes_added = 0

        self.extensions = set()
        self.extensions.add(".rom")
        self.extensions.add(".adf")
        self.extensions.add(".adz")
        self.extensions.add(".ipf")
        self.extensions.add(".dms")
        self.extensions.add(".bin")
        self.extensions.add(".cue")
        self.extensions.add(".iso")
        self.extensions.add(".wav")
        self.extensions.add(".zip")
        self.extensions.add(".lha")
        self.extensions.add(".rp9")
        self.extensions.add(".fs-uae")
        self.extensions.add(".7z")

        # if OGDClient.get_server() != "oagd.net":
        if True:
            self.extensions.add(".bin")

            # Amstrad CPC
            self.extensions.add(".dsk")

            # Atari 2600
            self.extensions.add(".a26")

            # Atari 5200
            self.extensions.add(".a52")

            # Atari 7800
            self.extensions.add(".a78")

            # Atari ST
            self.extensions.add(".st")  # disk images
            self.extensions.add(".img")  # bios images
            self.extensions.add(".bin")  # bios images

            # Game Boy
            self.extensions.add(".gb")

            # Game Boy Advance
            self.extensions.add(".gba")

            # Game Boy Color
            self.extensions.add(".gbc")

            # Master System
            self.extensions.add(".sms")

            # Mega Drive
            self.extensions.add(".md")
            self.extensions.add(".bin")
            self.extensions.add(".gen")

            # Nintendo
            self.extensions.add(".nes")

            # Super Nintendo
            self.extensions.add(".sfc")

            # ZX Spectrum
            self.extensions.add(".dsk")
            self.extensions.add(".tap")
            self.extensions.add(".tzx")
            self.extensions.add(".z80")

    def stop_check(self):
        if self._stop_check:
            return self._stop_check()

    def set_status(self, title, status):
        if self.on_status:
            self.on_status((title, status))

    def get_scan_dirs(self):
        return self.paths

    def purge_file_ids(self, file_ids):
        self.set_status(gettext("Scanning files"),
                        gettext("Purging old entries..."))
        database = FileDatabase.get_instance()
        for file_id in file_ids:
            database.delete_file(id=file_id)

    def scan(self):
        self.set_status(gettext("Scanning files"),
                        gettext("Scanning files"))
        database = FileDatabase.get_instance()
        # database.clear()
        scan_dirs = self.get_scan_dirs()
        
        if self.purge_other_dirs:
            all_database_file_ids = database.get_file_ids()
        else:
            all_database_file_ids = None

        for dir in scan_dirs:
            # this is important to make sure the database is portable across
            # operating systems
            dir = get_real_case(dir)

            self.database_file_ids = database.get_file_hierarchy_ids(dir)
            if self.purge_other_dirs:
                all_database_file_ids.difference_update(
                    self.database_file_ids)

            self.scan_dir(database, dir)

            # print("Remaining files:", self.database_file_ids)
            self.purge_file_ids(self.database_file_ids)

            self.set_status(gettext("Scanning files"),
                            gettext("Committing data..."))
            # update last_file_insert and last_file_delete
            database.update_last_event_stamps()
            print("FileScanner.scan - committing data")
            database.commit()

        if self.purge_other_dirs:
            self.purge_file_ids(all_database_file_ids)

        if self.stop_check():
            database.rollback()
        else:
            self.set_status(gettext("Scanning files"),
                            gettext("Committing data..."))
            print("FileScanner.scan - committing data")
            database.commit()

    def scan_dir(self, database, dir, all_files=False):
        if not os.path.exists(dir):
            return
        dir_content = os.listdir(dir)
        if not all_files:
            for name in dir_content:
                if not check_valid_name(name):
                    continue
                lname = name.lower()
                if lname.endswith(".slave") or lname.endswith(".slave"):
                    all_files = True
                    break

        for name in dir_content:
            if not check_valid_name(name):
                continue
            if self.stop_check():
                return
            if name in [".git", "Cache", "Save States"]:
                continue

            path = os.path.join(dir, name)
            if os.path.isdir(path):
                self.scan_dir(database, path, all_files=all_files)
                continue
            if not all_files:
                dummy, ext = os.path.splitext(path)
                ext = ext.lower()
                if ext not in self.extensions:
                    continue
            try:
                self.scan_file(database, path)
            except Exception:
                traceback.print_exc()

    def scan_file(self, database, path):
        name = os.path.basename(path)
        # path = os.path.normcase(os.path.normpath(path))

        self.scan_count += 1
        self.set_status(
            gettext("Scanning files ({count} scanned)").format(
                count=self.scan_count), name)

        try:
            st = os.stat(path)
        except:
            print("error stat-ing file", repr(path))
            return
        size = st.st_size
        mtime = int(st.st_mtime)

        result = database.find_file(path=path)
        # print(result)
        if result["path"]:
            if size == result["size"] and mtime == result["mtime"]:

                self.database_file_ids.remove(result["id"])
                # self.database_file_ids.difference_update(
                #         database.get_file_hierarchy_ids(path + "#"))
                
                # self.database_file_ids.difference_update(
                #         database.get_child_ids(id=result["id"]))
                return

        archive = Archive(path)
        # if archive.is_archive():

        file_id = self.scan_archive_stream(
            database, archive, path, name, size, mtime)

        for p in archive.list_files():
            if p.endswith("/"):
                # don't index archive directory entries
                continue
            # print(p)
            if self.stop_check():
                return
            # n = p.replace("\\", "/").replace("|", "/").split("/")[-1]
            n = os.path.basename(p)
            self.scan_count += 1
            self.scan_archive_stream(
                database, archive, p, n, size, mtime, parent=file_id)
        if self.stop_check():
            return

    def scan_archive_stream(
            self, database, archive, path, name, size, mtime, parent=None):
        self.set_status(
            gettext("Scanning files ({count} scanned)").format(
                count=self.scan_count), name)
        base_name, ext = os.path.splitext(name)
        ext = ext.lower()

        f = archive.open(path)
        s = hashlib.sha1()
        while True:
            if self.stop_check():
                return
            data = f.read(65536)
            if not data:
                break
            s.update(data)
        sha1 = s.hexdigest()

        if ext == ".rom":
            try:
                sha1_dec = ROMManager.decrypt_archive_rom(archive, path)["sha1"]
            except Exception:
                import traceback
                traceback.print_exc()
                sha1_dec = None
            if sha1_dec:
                if sha1_dec != sha1:
                    print("found encrypted rom {0} => {1}".format(
                        sha1, sha1_dec))
                    # sha1 is now the decrypted sha1, not the actual sha1 of the
                    # file itself, a bit ugly, since md5 and crc32 are still
                    # encrypted hashes, but it works well with the kickstart
                    # lookup mechanism
                    sha1 = sha1_dec
            else:
                # If the ROM was encrypted and could not be decrypted, we
                # don't add it to the database. This way, the file will be
                # correctly added on later scans if rom.key is added to the
                # directory.
                return None

        if parent:
            path = "#/" + path.rsplit("#/", 1)[1]
        file_id = database.add_file(
            path=path, sha1=sha1, mtime=mtime, size=size, parent=parent)

        self.files_added += 1
        self.bytes_added += size

        if ext == '.rom':
            if self.on_rom_found:
                self.on_rom_found(path, sha1)
        return file_id


def check_valid_name(name):
    # check that the file is actually unicode object (indirectly). listdir
    # will return str objects for file names on Linux with invalid encoding.
    # FIXME: not needed for Python 3
    try:
        str(name)
    except UnicodeDecodeError:
        return False
    return True



from weakref import WeakValueDictionary
from fs_uae_workspace.vfs import get_vfs_item


window_uris = WeakValueDictionary()
# window_uris = {}
opened_count = 0


def set_initial_window_position(window):
    global opened_count
    x = 100 + (opened_count % 10) * 40
    y = x
    window.set_position((x, y))
    opened_count += 1


def raise_window(uri):
    print("raise_window", uri)
    try:
        window = window_uris[uri]
    except KeyError:
        return False
    # window.raise_()
    # # window.setFocus()
    # window.activateWindow()
    window.raise_and_activate()
    return True


def register_window(uri, window):
    print("register_window", uri, window)

    # def on_window_closed():
    #     print("on_window_closed")
    #     del window_uris[uri]
    # window.closed.connect(on_window_closed)

    window_uris[uri] = window


def shell_open(uri, args=None, parent=None, center=None):
    if args is None:
        args = []
    item = get_vfs_item(uri)
    import fsui
    if parent:
        fsui.default_window_parent.append(parent)
    if center:
        fsui.default_window_center.append(center)
    try:
        item.open(args)
    finally:
        if parent:
            fsui.default_window_parent.pop()
        if center:
            fsui.default_window_center.pop()


# noinspection PyPep8Naming
def SimpleApplication(window_class):

    def application(uri, args):
        fake_uri = uri + repr(args)
        if not raise_window(fake_uri):
            window = window_class()
            window.show()
            register_window(fake_uri, window)

    return application


# noinspection PyUnresolvedReferences
from .window import Window, QWindow



'''OpenGL extension INTEL.swap_event

This module customises the behaviour of the 
OpenGL.raw.GLX.INTEL.swap_event to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension adds a new event type, GLX_BUFFER_SWAP_COMPLETE_INTEL,
	which is sent to the client via the X11 event stream and
	selected/consumed by the normal GLX event mask mechanisms, to indicate
	when a previously queued swap has completed.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/INTEL/swap_event.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper

import ctypes
from OpenGL.raw.GLX import _types
from OpenGL.raw.GLX.INTEL.swap_event import *
from OpenGL.raw.GLX.INTEL.swap_event import _EXTENSION_NAME

def glInitSwapEventINTEL():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


from fsbc.util import memoize


class Platform(object):

    pass


class PlatformHandler(object):

    def __init__(self):
        pass

    @classmethod
    def create(cls, platform_id):
        class_ = cls.get_platform_class(platform_id)
        return class_()

    @classmethod
    def get_platform_class(cls, platform_id):
        platform_id = platform_id.lower()
        try:
            return platforms[platform_id]
        except KeyError:
            return UnsupportedPlatformHandler

    @classmethod
    @memoize
    def get_platform_name(cls, platform_id):
        return cls.get_platform_class(platform_id).PLATFORM_NAME

    @classmethod
    @memoize
    def get_platform_ids(cls):
        return sorted(platforms.keys())


class UnsupportedPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Unsupported"


from .platforms.amiga import AmigaPlatformHandler
from .platforms.amstrad_cpc import AmstradCPCPlatformHandler
from .platforms.arcade import ArcadePlatformHandler
from .platforms.atari_2600 import Atari2600PlatformHandler
from .platforms.atari_5200 import Atari5200PlatformHandler
from .platforms.atari_7800 import Atari7800PlatformHandler
from .platforms.atari_st import AtariSTPlatformHandler
from .platforms.cdtv import CDTVPlatformHandler
from .platforms.cd32 import CD32PlatformHandler
from .platforms.commodore_64 import Commodore64PlatformHandler
from .platforms.dos import DOSPlatformHandler
from .platforms.game_boy import GameBoyPlatformHandler
from .platforms.game_boy_advance import GameBoyAdvancePlatformHandler
from .platforms.game_boy_color import GameBoyColorPlatformHandler
from .platforms.game_gear import GameGearPlatformHandler
from .platforms.lynx import LynxPlatformHandler
from .platforms.master_system import MasterSystemPlatformHandler
from .platforms.mega_drive import MegaDrivePlatformHandler
from .platforms.nintendo import NintendoPlatformHandler
from .platforms.super_nintendo import SuperNintendoPlatformHandler
from .platforms.turbografx_16 import TurboGrafx16PlatformHandler
from .platforms.zx_spectrum import ZXSpectrumPlatformHandler


platforms = {
    "amiga": AmigaPlatformHandler,
    "amstrad-cpc": AmstradCPCPlatformHandler,
    "arcade": ArcadePlatformHandler,
    "atari-2600": Atari2600PlatformHandler,
    "atari-5200": Atari5200PlatformHandler,
    "atari-7800": Atari7800PlatformHandler,
    "atari-st": AtariSTPlatformHandler,
    "cd32": CD32PlatformHandler,
    "cdtv": CDTVPlatformHandler,
    "commodore-64": Commodore64PlatformHandler,
    "dos": DOSPlatformHandler,
    "game-boy": GameBoyPlatformHandler,
    "game-boy-advance": GameBoyAdvancePlatformHandler,
    "game-boy-color": GameBoyColorPlatformHandler,
    "game-gear": GameGearPlatformHandler,
    "lynx": LynxPlatformHandler,
    "master-system": MasterSystemPlatformHandler,
    "mega-drive": MegaDrivePlatformHandler,
    "nintendo": NintendoPlatformHandler,
    "super-nintendo": SuperNintendoPlatformHandler,
    "turbografx-16": TurboGrafx16PlatformHandler,
    "zx-spectrum": ZXSpectrumPlatformHandler,
}



import os
import sys
import hashlib
import traceback
from collections import defaultdict
from fs_uae_launcher.Options import Option, Options
from fs_uae_launcher.ui.config.HardDriveGroup import HardDriveGroup
from fs_uae_launcher.ui.download import DownloadGameWindow, DownloadTermsDialog
from fsgs.Downloader import Downloader
from fs_uae_launcher.PluginHelper import PluginHelper
from fsbc.Application import app
from configparser import ConfigParser, NoSectionError
from fsbc.system import windows, macosx
from fsbc.task import Task
import fsbc.user
from fsbc.util import unused, is_uuid
from fsgs.application import ApplicationMixin
from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.platform import PlatformHandler
import fstd.desktop
import fsbc.fs as fs
import fsui as fsui
from fsgs.context import fsgs
from fsgs.Database import Database
from .ui.MainWindow import MainWindow
from fsgs.FileDatabase import FileDatabase
from fsbc.Paths import Paths
from fsgs.amiga.Amiga import Amiga
from fsgs.amiga.ROMManager import ROMManager
from .Config import Config
from .ConfigurationScanner import ConfigurationScanner
from fsgs.FSGSDirectories import FSGSDirectories
from .I18N import gettext, initialize_locale
from .Settings import Settings
from .UpdateManager import UpdateManager


class FSUAELauncher(ApplicationMixin, fsui.Application):

    def __init__(self):
        fsui.Application.__init__(self, "fs-uae-launcher")
        self.set_icon(fsui.Icon("fs-uae-launcher", "pkg:fs_uae_launcher"))
        if fsui.use_qt:
            initialize_qt_style(self.qapplication)

    @staticmethod
    def get_game_database_path():
        launcher_dir = FSGSDirectories.get_launcher_dir()
        path = os.path.join(launcher_dir, "Game Database.sqlite")
        return path

    # noinspection PyMethodMayBeStatic
    def on_idle(self):
        fsgs.signal.process()

    def start(self):
        print("FSUAELauncherApplication.start")

        # GameDatabase.set_database_path(self.get_game_database_path())
        # OverlayDatabase.set_database_path(self.get_overlay_database_path())

        self.parse_arguments()
        self.load_settings()

        language = Settings.get("language")
        initialize_locale(language)

        self.config_startup_scan()
        self.kickstart_startup_scan()

        # sys.exit(1)

        # FIXME: should now sanitize check some options -for instance,
        # - check if configured joysticks are still connected
        # - check if paths still exists, etc

        # Config.update_kickstart()

        icon = None

        def check_icon(path):
            path = os.path.join(path, "fs-uae-launcher.ico")
            if os.path.exists(path):
                return path
            return None

        if not icon:
            icon = check_icon("share/fs-uae-launcher")
        if not icon:
            icon = check_icon("launcher/share/fs-uae-launcher")
        # FIXME: should check data directories (XDG_DATA_DIRS) properly
        # instead
        if not icon:
            # this encoding / decoding is a bit ridiculous, but, this is
            # for Python 2.x..
            icon = check_icon(os.path.expanduser(
                "~/.local/share/fs-uae-launcher".encode(
                    sys.getfilesystemencoding())).decode(
                sys.getfilesystemencoding()))
        if not icon:
            icon = check_icon("/usr/local/share/fs-uae-launcher")
        if not icon:
            icon = check_icon("/usr/share/fs-uae-launcher")
        if macosx:
            # Icons come from the app bundles
            icon = None

        plugins_dir = FSGSDirectories.get_plugins_dir()
        if plugins_dir:
            dont_write_bytecode = sys.dont_write_bytecode
            try:
                self.load_plugins(plugins_dir)
            finally:
                sys.dont_write_bytecode = dont_write_bytecode

        config_path = None
        archive_path = None
        floppy_image_path = None
        config_uuid = None

        # FIXME: replace argument "parsing" with use of argparse module
        # at some point

        if sys.argv[-1].endswith(".fs-uae"):
            config_path = sys.argv[-1]
        elif sys.argv[-1].endswith(".zip"):
            archive_path = sys.argv[-1]
        elif sys.argv[-1].endswith(".lha"):
            archive_path = sys.argv[-1]
        elif sys.argv[-1].endswith(".adf"):
            # FIXME: support / check for other disk images as well
            floppy_image_path = sys.argv[-1]
        elif is_uuid(sys.argv[-1]):
            config_uuid = sys.argv[-1].lower()

        if config_path:
            print("config path given:", config_path)
            if not os.path.exists(config_path):
                print("config path does not exist", file=sys.stderr)
                return False
            Config.load_file(config_path)
            Settings.set("parent_uuid", "")
            self.start_game()
            return False

        elif archive_path:
            print("archive path given:", archive_path)
            if not os.path.exists(archive_path):
                print("archive path does not exist", file=sys.stderr)
                return False
            # FIXME: if archive is determined to contain (only) floppy disk
            # images, load those into floppy drives instead
            values = HardDriveGroup.generate_config_for_archive(archive_path)
            values["hard_drive_0"] = archive_path
            Config.load(values)
            Settings.set("parent_uuid", "")
            self.start_game()
            return False

        elif floppy_image_path:
            values = {
                "floppy_drive_0": floppy_image_path,
            }
            Config.load(values)
            Settings.set("parent_uuid", "")
            self.start_game()
            return False

        elif config_uuid:
            print("config uuid given:", config_uuid)
            variant_uuid = config_uuid
            # values = fsgs.game.set_from_variant_uuid(variant_uuid)
            if fsgs.load_game_variant(variant_uuid):
                print("loaded variant")
            else:
                print("could not load variant, try to load game")
                game_uuid = config_uuid
                variant_uuid = fsgs.find_preferred_game_variant(game_uuid)
                print("preferred variant:", variant_uuid)
                fsgs.load_game_variant(variant_uuid)
            self.start_game()
            return False

        window = MainWindow(fsgs, icon=icon)
        MainWindow.instance = window
        window.show()

        if "--workspace" in sys.argv:
            window.set_position((300, 200))
            from fs_uae_workspace.desktop import get_desktop_window
            get_desktop_window()

        UpdateManager.run_update_check()
        return True

    @staticmethod
    def load_plugins(plugins_dir):
        print("loading plugins")
        for full_name in os.listdir(plugins_dir):
            name, ext = os.path.splitext(full_name)
            if ext.lower() != ".py":
                continue
            path = os.path.join(plugins_dir, full_name)
            print("loading", path)
            name = "plugin_" + hashlib.sha1(path.encode("UTF-8")).hexdigest()
            print(name)
            import imp
            try:
                # with open(path, "r") as f:
                #     plugin = imp.load_source(name, name, f)
                plugin = imp.load_source(name, path)
                print("name:", getattr(plugin, "name", ""))
                print("version:", getattr(plugin, "version", ""))
                plugin.fsgs = fsgs
                plugin.plugin_dir = os.path.dirname(path)
                plugin.fsgs_init()
            except Exception:
                traceback.print_exc()
                continue
            else:
                print(name, "initialized")

    def load_settings(self):
        path = app.get_settings_path()
        print("loading last config from " + repr(path))
        if not os.path.exists(path):
            print("settings file does not exist")
        # noinspection PyArgumentList
        cp = ConfigParser(interpolation=None)
        try:
            cp.read([path])
        except Exception as e:
            print(repr(e))
            return

        config = {}
        try:
            keys = cp.options("config")
        except NoSectionError:
            keys = []
        for key in keys:
            config[key] = fs.from_utf8_str(cp.get("config", key))
        for key, value in config.items():
            print("loaded", key, value)
            fsgs.config.values[key] = value

    def parse_arguments(self):
        pass
        # for arg in sys.argv:
        #     if arg.startswith("--"):
        #         if "=" in arg:
        #             key, value = arg[2:].split("=", 1)
        #             key = key.replace("-", "_")
        #             if key == "base_dir":
        #                 Settings.set("base_dir", value)

    def save_settings(self):
        extra = {}
        for key, value in fsgs.config.values.items():
            if key.startswith("__"):
                # keys starting with __ are never saved
                continue
            extra["config/" + str(key)] = str(value)
        app.settings.save(extra=extra)

    @staticmethod
    def get_dir_mtime_str(path):
        try:
            return str(int(os.path.getmtime(path)))
        except Exception:
            return "0"

    def config_startup_scan(self):
        configs_dir = FSGSDirectories.get_configurations_dir()
        print("config_startup_scan", configs_dir)
        print(Settings.settings)

        settings_mtime = Settings.get("configurations_dir_mtime")
        dir_mtime = self.get_dir_mtime_str(configs_dir)
        if settings_mtime == dir_mtime + "+" + str(Database.VERSION):
            print("... mtime not changed", settings_mtime, dir_mtime)
            return

        database = Database.get_instance()
        file_database = FileDatabase.get_instance()

        print("... database.find_local_configurations")
        local_configs = Database.get_instance().find_local_configurations()
        print("... walk configs_dir")
        for dir_path, dir_names, file_names in os.walk(configs_dir):
            for file_name in file_names:
                if not file_name.endswith(".fs-uae"):
                    continue
                path = Paths.join(dir_path, file_name)
                if path in local_configs:
                    local_configs[path] = None
                    # already exists in database
                    continue
                name, ext = os.path.splitext(file_name)
                # search = ConfigurationScanner.create_configuration_search(
                # name)
                scanner = ConfigurationScanner()
                print("[startup] adding config", path)
                file_database.delete_file(path=path)
                with open(path, "rb") as f:
                    sha1 = hashlib.sha1(f.read()).hexdigest()
                file_database.add_file(path=path, sha1=sha1)

                game_id = database.add_game(
                    path=path, name=scanner.create_configuration_name(name))
                database.update_game_search_terms(
                    game_id, scanner.create_search_terms(name))

        for path, id in local_configs.items():
            if id is not None:
                print("[startup] removing configuration", path)
                database.delete_game(id=id)
                file_database.delete_file(path=path)
        print("... commit")
        database.commit()

        Settings.set(
            "configurations_dir_mtime",
            self.get_dir_mtime_str(configs_dir) + "+" + str(Database.VERSION))

    def kickstart_startup_scan(self):
        print("kickstart_startup_scan")
        kickstarts_dir = FSGSDirectories.get_kickstarts_dir()
        if Settings.get("kickstarts_dir_mtime") == \
                self.get_dir_mtime_str(kickstarts_dir):
            print("... mtime not changed")
        else:
            file_database = FileDatabase.get_instance()
            print("... database.find_local_roms")
            local_roms = file_database.find_local_roms()
            print("... walk kickstarts_dir")
            for dir_path, dir_names, file_names in os.walk(kickstarts_dir):
                for file_name in file_names:
                    if not file_name.lower().endswith(".rom") and not \
                            file_name.lower().endswith(".bin"):
                        continue
                    path = Paths.join(dir_path, file_name)
                    if path in local_roms:
                        local_roms[path] = None
                        # already exists in database
                        continue
                    print("[startup] adding kickstart", path)
                    ROMManager.add_rom_to_database(path, file_database)
            print(local_roms)
            for path, id in local_roms.items():
                if id is not None:
                    print("[startup] removing kickstart", path)
                    file_database.delete_file(id=id)
            print("... commit")
            file_database.commit()
            Settings.set(
                "kickstarts_dir_mtime",
                self.get_dir_mtime_str(kickstarts_dir))

        amiga = Amiga.get_model_config("A500")
        for sha1 in amiga["kickstarts"]:
            if fsgs.file.find_by_sha1(sha1=sha1):
                break
        else:
            file_database = FileDatabase.get_instance()
            self.amiga_forever_kickstart_scan()
            file_database.commit()

    def amiga_forever_kickstart_scan(self):
        if windows:
            print("amiga forever kickstart scan")
            path = fsbc.user.get_common_documents_dir()
            path = os.path.join(path, "Amiga Files", "Shared", "rom")
            self.scan_dir_for_kickstarts(path)

    @staticmethod
    def scan_dir_for_kickstarts(scan_dir):
        file_database = FileDatabase.get_instance()
        for dir_path, dir_names, file_names in os.walk(scan_dir):
            for file_name in file_names:
                if not file_name.endswith(".rom"):
                    continue
                path = Paths.join(dir_path, file_name)
                if file_database.find_file(path=path):
                    continue
                print("[startup] adding kickstart", path)
                ROMManager.add_rom_to_database(path, file_database)

    @classmethod
    def start_game(cls):
        from .netplay.Netplay import Netplay
        if Netplay.game_channel:
            Netplay.start_netplay_game()
        else:
            cls.start_local_game()

    @classmethod
    def start_local_game(cls):
        print("START LOCAL GAME")
        print("x_missing_files", Config.get("x_missing_files"))

        if Config.get("x_missing_files"):
            if Config.get("download_file"):
                if Config.get("download_terms") and not \
                        Downloader.check_terms_accepted(
                        Config.get("download_file"),
                        Config.get("download_terms")):
                    from .ui.MainWindow import MainWindow
                    dialog = DownloadTermsDialog(MainWindow.instance, fsgs)
                    if not dialog.show_modal():
                        return

            elif Config.get("download_page"):
                from .ui.MainWindow import MainWindow
                # fsui.show_error(_("This game must be downloaded first."))
                DownloadGameWindow(MainWindow.instance, fsgs).show()
                return
            else:
                fsui.show_error(
                    gettext("This game variant cannot be started "
                            "because you don't have all required files."))
                return

        platform_id = Config.get("platform")
        amiga_platform = platform_id in ["", "amiga", "cdtv", "cd32"]
        if amiga_platform:
            cls.start_local_game_amiga()
        else:
            cls.start_local_game_other()

    @classmethod
    def start_local_game_other(cls):
        # platform_id = Config.get("platform")
        # FIXME:
        # if platform_id == "super-nintendo":
        #     platform_id = "snes"
        database_name = Config.get("__database")
        variant_uuid = Config.get("variant_uuid")
        assert variant_uuid

        fsgs.game.set_from_variant_uuid(database_name, variant_uuid)
        platform_handler = PlatformHandler.create(fsgs.game.platform.id)
        runner = platform_handler.get_runner(fsgs)

        task = RunnerTask(runner)
        from .ui.LaunchDialog import LaunchDialog
        from .ui.MainWindow import MainWindow
        dialog = LaunchDialog(
            MainWindow.instance, gettext("Launching Game"), task)
        task.start()
        dialog.show_modal()
        dialog.close()

    @classmethod
    def start_local_game_amiga(cls):
        # make sure x_kickstart_file is initialized
        Config.set_kickstart_from_model()

        # if not Config.get("x_kickstart_file"):  # or not \
        #     #  os.path.exists(Config.get("kickstart_file")):
        #     fsui.show_error(
        #         gettext("No kickstart found for this model. Use the 'Import "
        #                 "Kickstarts' function from the menu."))
        #     return
        cs = Amiga.get_model_config(Config.get("amiga_model"))["ext_roms"]
        if len(cs) > 0:
            # extended kickstart ROM is needed
            if not Config.get("x_kickstart_ext_file"):
                fsui.show_error(
                    gettext("No extended kickstart found for this model. "
                            "Try 'scan' function."))
                return

        config = Config.copy()
        prepared_config = cls.prepare_config(config)

        model = Config.get("amiga_model")
        if model.startswith("CD32"):
            platform = "CD32"
        elif model == "CDTV":
            platform = "CDTV"
        else:
            platform = "Amiga"
        name = Settings.get("config_name")
        uuid = Config.get("x_game_uuid")
        
        from fsgs.SaveStateHandler import SaveStateHandler
        save_state_handler = SaveStateHandler(fsgs, name, platform, uuid)

        from fsgs.amiga.LaunchHandler import LaunchHandler
        launch_handler = LaunchHandler(fsgs, name, prepared_config,
                                       save_state_handler)

        from .ui.LaunchDialog import LaunchDialog
        from .ui.MainWindow import MainWindow
        task = AmigaLaunchTask(launch_handler)
        # dialog = LaunchDialog(MainWindow.instance, launch_handler)
        dialog = LaunchDialog(
            MainWindow.instance, gettext("Launching FS-UAE"), task)

        def on_show_license_information(license_text):
            unused(license_text)
            # FIXME: don't depend on wx here
            # noinspection PyUnresolvedReferences
            # import wx
            # license_dialog = wx.MessageDialog(
            #     dialog, license_text, _("Terms of Use"),
            #     wx.OK | wx.CANCEL | wx.CENTRE)
            # license_dialog.CenterOnParent()
            # result = license_dialog.ShowModal()
            # return result == wx.ID_OK
            # FIXME
            return True
        fsgs.file.on_show_license_information = on_show_license_information

        task.start()
        dialog.show_modal()
        dialog.close()

    @classmethod
    def prepare_config(cls, original_config):
        config = defaultdict(str)
        for key, value in app.settings.values.items():
            if key in Config.config_keys:
                print("... ignoring config key from settings:", key)
                continue
            config[key] = value

        config["base_dir"] = FSGSDirectories.get_base_dir()

        for key, value in original_config.items():
            config[key] = value

        if not config["joystick_port_0_mode"]:
            config["joystick_port_0_mode"] = "mouse"
        if not config["joystick_port_1_mode"]:
            if config["amiga_model"].startswith("CD32"):
                config["joystick_port_1_mode"] = "cd32 gamepad"
            else:
                config["joystick_port_1_mode"] = "joystick"
        if not config["joystick_port_2_mode"]:
            config["joystick_port_2_mode"] = "none"
        if not config["joystick_port_3_mode"]:
            config["joystick_port_3_mode"] = "none"

        from .DeviceManager import DeviceManager
        devices = DeviceManager.get_devices_for_ports(config)
        for port in range(4):
            key = "joystick_port_{0}".format(port)
            if not config.get(key):
                # key not set, use calculated default value
                config[key] = devices[port].id

        for remove_key in ["database_username", "database_password",
                           "database_username", "database_email",
                           "database_auth", "device_id"]:
            if remove_key in config:
                del config[remove_key]

        # overwrite netplay config

        if config.get("__netplay_host", ""):
            config["netplay_server"] = config["__netplay_host"]
        if config.get("__netplay_password", ""):
            config["netplay_password"] = config["__netplay_password"]
        if config.get("__netplay_port", ""):
            config["netplay_port"] = config["__netplay_port"]

        # copy actual kickstart options from x_ options

        config["kickstart_file"] = config["x_kickstart_file"]
        config["kickstart_ext_file"] = config["x_kickstart_ext_file"]

        if not config["kickstart_file"]:
            # Warning will have been shown on the status bar
            config["kickstart_file"] = "internal"

        # Copy default configuration values from model defaults. The main
        # purpose of this is to let the launch code know about implied defaults
        # so it can for example configure correct ROM files for expansions.

        model_config = Amiga.get_current_config(config)
        for key, value in model_config["defaults"].items():
            if not config.get(key):
                config[key] = value

        # make sure FS-UAE does not load other config files (Host.fs-uae)
        config["end_config"] = "1"

        if config.get("__netplay_game", ""):
            print("\nfixing config for netplay game")
            for key in [x for x in config.keys() if x.startswith("uae_")]:
                print("* removing option", key)
                del config[key]

        return config


def initialize_qt_style(qapplication):
    from fsui.qt import QStyleFactory, QPalette, QColor, Qt, QFont
    fusion_variant = ""

    launcher_theme = Settings.get("launcher_theme")
    if launcher_theme == "standard":
        use_fusion_theme = False
    elif launcher_theme == "native":
        # native is an older deprecated name for standard
        use_fusion_theme = False
    elif launcher_theme == "fusion-adwaita":
        use_fusion_theme = True
        fusion_variant = "adwaita"
    elif launcher_theme == "fusion-plain":
        use_fusion_theme = True
    elif launcher_theme == "fusion-dark":
        use_fusion_theme = True
        fusion_variant = "dark"
    else:
        use_fusion_theme = True
        if fstd.desktop.is_running_gnome_3():
            fusion_variant = "adwaita"

    if "--launcher-theme=fusion-dark" in sys.argv:
        use_fusion_theme = True
        fusion_variant = "dark"

    font = qapplication.font()
    print("FONT: Default is {} {}".format(font.family(), font.pointSize()))
    Options.get(Option.LAUNCHER_FONT_SIZE)["default"] = font.pointSize()

    if use_fusion_theme:
        # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
        qapplication.setStyle(QStyleFactory.create("Fusion"))
        if fusion_variant == "adwaita":
            pa = QPalette()
            pa.setColor(QPalette.Window, QColor(237, 237, 237))
            pa.setColor(QPalette.AlternateBase, QColor(237, 237, 237))
            pa.setColor(QPalette.Button, QColor(237, 237, 237))
            qapplication.setPalette(pa)
        elif fusion_variant == "dark":
            pa = QPalette()
            pa.setColor(QPalette.Window, QColor(53, 53, 53))
            pa.setColor(QPalette.WindowText, Qt.white)
            pa.setColor(QPalette.Base, QColor(25, 25, 25))
            pa.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
            pa.setColor(QPalette.ToolTipBase, Qt.white)
            pa.setColor(QPalette.ToolTipText, Qt.white)
            pa.setColor(QPalette.Text, Qt.white)
            pa.setColor(QPalette.Button, QColor(53, 53, 53))
            pa.setColor(QPalette.ButtonText, Qt.white)
            pa.setColor(QPalette.BrightText, Qt.red)
            pa.setColor(QPalette.Link, QColor(42, 130, 218))
            pa.setColor(QPalette.Highlight, QColor(42, 130, 218))
            pa.setColor(QPalette.HighlightedText, Qt.black)
            qapplication.setPalette(pa)
            qapplication.setStyleSheet(
                "QToolTip { color: #ffffff; background-color: #2a82da; "
                "border: 1px solid white; }")

        try:
            launcher_font_size = int(Settings.get("launcher_font_size"))
        except ValueError:
            launcher_font_size = 0
        if launcher_font_size:
            print("FONT: Override size {}".format(launcher_font_size))
            font.setPointSize(launcher_font_size)
            qapplication.setFont(font)

    plugin_helper = PluginHelper()
    for res in plugin_helper.find_resource_dirs(
            "fs-uae-launcher-theme"):
        qt_css = os.path.join(res, "stylesheet.qss")
        if os.path.exists(qt_css):
            with open(qt_css, "rb") as f:
                data = f.read()
            qapplication.setStyleSheet(data)


# FIXME: Files to clean up:
# Documents/FS-UAE/Cache/File Database.sqlite
# Documents/FS-UAE/Data/Game Database.sqlite
# Documents/FS-UAE/Data/Launcher.sqlite
# Documents/FS-UAE/Logs/FS-UAE.log
# Documents/FS-UAE/Logs/FS-UAE.log.txt


class AmigaLaunchTask(Task):

    def __init__(self, launch_handler):
        Task.__init__(self, "Amiga Launch Task")
        self.launch_handler = launch_handler

    def run(self):
        self.launch_handler.run_sequence()


class RunnerTask(Task):

    def __init__(self, runner):
        Task.__init__(self, "Runner Task")
        self.runner = runner

    def run(self):
        device_helper = EnumerateHelper()
        device_helper.default_port_selection(self.runner.ports)

        self.runner.prepare()
        process = self.runner.run()
        process.wait()
        self.runner.finish()



from fs_uae_launcher.Settings import Settings
import fsui as fsui
from fsui.extra.iconheader import NewIconHeader
from ...I18N import gettext
from .OptionUI import OptionUI


class VideoSyncSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("video-settings", "pkg:fs_uae_workspace"),
            gettext("Video Synchronization Settings"),
            gettext("Synchronize FS-UAE with your display for smooth video"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            group = OptionUI.create_group(self, name)
            self.layout.add(group, fill=True,
                            margin_top=10, margin_bottom=10)
            return group

        # label = fsui.MultiLineLabel(self, gettext(
        #     "Enabling the following option will prevent tearing from "
        #     "occurring, but will also use more CPU. Input latency "
        #     "may become slightly higher. If you encounter slowdowns "
        #     "(esp. in Linux desktop environments with compositing "
        #     "enabled), disable this option or follow the link at "
        #     "the bottom to learn more."), 640)
        # self.layout.add(label, fill=True, margin_top=0)

        label = fsui.MultiLineLabel(self, gettext(
            "Enabling the following option will prevent tearing from "
            "occurring, but will also use more CPU. Input latency "
            "may become slightly higher."), 640)
        self.layout.add(label, fill=True, margin_top=0)

        # add_option("video_sync")
        self.vblank_checkbox = fsui.HeadingCheckBox(self, gettext(
            "Synchronize buffer swaps with display (prevents tearing)"))
        self.layout.add(self.vblank_checkbox, margin_top=20)

        # label = fsui.MultiLineLabel(self, gettext(
        #     ""), 640)
        # self.layout.add(label, fill=True, margin_top=10)

        # hori_layout = fsui.HorizontalLayout()
        # self.layout.add(hori_layout, fill=True, margin_top=10)
        # hori_layout.add(fsui.ImageView(self, fsui.Image(
        #     "fs_uae_launcher:res/16/world_link.png")))
        # label = fsui.URLLabel(self, gettext(
        #     "Read more about synchronizing FS-UAE with the display"),
        #     "http://fs-uae.net/video-sync")
        # hori_layout.add(label, margin_left=6)

        self.sync_method_label = fsui.MultiLineLabel(self, gettext(
            "Depending on your OS and OpenGL drivers, synchronizing "
            "can use needlessly much CPU (esp. applies to "
            "Linux). You can experiment with different sync methods "
            "to improve performance."), 640)
        self.layout.add(self.sync_method_label, fill=True, margin_top=20)

        self.sync_method_group = add_option("video_sync_method")

        self.smooth_label = fsui.MultiLineLabel(self, gettext(
            "In order to get really smooth Amiga graphics, you need to "
            "enable the following option, and also make sure your display "
            "is running at 50Hz (for PAL) or 60Hz (for NTSC)."), 640)
        self.layout.add(self.smooth_label, fill=True, margin_top=20)

        self.full_sync_checkbox = fsui.HeadingCheckBox(self, gettext(
            "Also synchronize emulation with display when possible "
            "(smooth scrolling)"))
        self.layout.add(self.full_sync_checkbox, margin_top=20)

        self.low_latency_group = add_option("low_latency_vsync")

        self.layout.add_spacer(0, expand=True)

        # label = fsui.MultiLineLabel(self, gettext(
        #     "If you encounter slowdowns or increased stuttering when "
        #     "synchronization is enabled (esp. in Linux desktop "
        #     "environments with compositing enabled), keep these options "
        #     "disabled or learn more."), 640)
        # self.layout.add(label, fill=True, margin_top=20)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=10)
        hori_layout.add(fsui.ImageView(self, fsui.Image(
            "fs_uae_launcher:res/16/world_link.png")))
        label = fsui.URLLabel(self, gettext(
            "How to achieve perfectly smooth scrolling"),
            "http://fs-uae.net/perfectly-smooth-scrolling")
        hori_layout.add(label, margin_left=6)

        # hori_layout = fsui.HorizontalLayout()
        # self.layout.add(hori_layout, fill=True, margin_top=4)
        # hori_layout.add(fsui.ImageView(self, fsui.Image(
        #     "fs_uae_launcher:res/16/world_link.png")))
        # label = fsui.URLLabel(self, gettext(
        #     "Read more about synchronizing FS-UAE with the display"),
        #     "http://fs-uae.net/video-sync")
        # hori_layout.add(label, margin_left=6)

        label = fsui.MultiLineLabel(self, gettext(
            "Synchronizing with the display can in some cases cause "
            "increased stuttering and low frame rates (esp. in some Linux "
            "desktop environments with compositing enabled).") +
            " <a href='http://fs-uae.net/video-synchronization-issues'>"
            "{0}</a>.".format(
            gettext("Read more")), 640)
        self.layout.add(label, fill=True, margin_top=20)

        Settings.add_listener(self)
        for key in ["video_sync"]:
            self.on_setting(key, Settings.get(key))

        self.vblank_checkbox.changed.connect(self.on_vblank_changed)
        self.full_sync_checkbox.changed.connect(self.on_full_sync_changed)

    def on_destroy(self):
        Settings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "video_sync":
            value = value.lower()
            self.vblank_checkbox.check(value in ["auto", "full", "vblank"])
            self.full_sync_checkbox.check(value in ["auto", "full"])
            self.update_widgets()
        # elif key == "video_sync_method":
        #     pass

    def on_vblank_changed(self):
        self.on_either_checkbox_changed()

    def on_full_sync_changed(self):
        self.on_either_checkbox_changed()

    def on_either_checkbox_changed(self):
        value = ""
        if self.vblank_checkbox.is_checked():
            if self.full_sync_checkbox.is_checked():
                value = "auto"
            else:
                value = "vblank"
        Settings.set("video_sync", value)

    def update_widgets(self):
        value = Settings.get("video_sync")
        vblank = value in ["vblank", "auto", "full"]
        full = value in ["auto", "full"]

        self.full_sync_checkbox.enable(vblank)
        self.sync_method_group.label.enable(vblank)
        self.sync_method_group.widget.enable(vblank)
        self.sync_method_group.help_button.enable(vblank)
        self.sync_method_label.enable(vblank)
        self.smooth_label.enable(vblank)
        self.low_latency_group.label.enable(full)
        self.low_latency_group.widget.enable(full)
        self.low_latency_group.help_button.enable(full)



from fsui.qt import QDesktopWidget, QMainWindow
try:
    from Xlib import X
    from Xlib.display import Display
    from Xlib.error import BadAccess
    from Xlib.protocol.request import InternAtom
except ImportError:
    X = None
    Display = None
    BadAccess = None
    InternAtom = None


class RealDesktopWindow(QMainWindow):

    def __init__(self):
        # self.widget = QWidget()
        self.widget = None
        # flags = 0
        # flags |= Qt.FramelessWindowHint
        # flags |= Qt.Desktop
        # flags |= Qt.WindowStaysOnBottomHint
        # flags |= Qt.CustomizeWindowHint
        # flags |= Qt.WindowCloseButtonHint
        # flags |= Qt.WindowMinimizeButtonHint
        # flags |= Qt.Tool
        # flags |= Qt.WindowDoesNotAcceptFocus
        super().__init__(self.widget)
        self.setWindowTitle("FS-UAE Workspace")
        g = QDesktopWidget().screenGeometry()
        self.setFixedSize(g.width(), g.height())
        self.move(0, 0)

        # self.setFocusPolicy(Qt.NoFocus)
        # self.setAttribute(Qt.WA_ShowWithoutActivating)
        self.set_x11_properties()

    # noinspection PyPep8Naming
    def set_x11_properties(self):
        xid = self.winId()
        display = Display()
        window = display.create_resource_object("window", xid)
        # _NET_WM_DESKTOP = display.intern_atom("_NET_WM_DESKTOP")
        _NET_WM_STRUT = display.intern_atom("_NET_WM_STRUT")
        _NET_WM_STATE = display.intern_atom("_NET_WM_STATE")
        ATOM = display.intern_atom("ATOM")
        # _NET_WM_STATE_SKIP_TASKBAR = display.intern_atom(
        #     "_NET_WM_STATE_SKIP_TASKBAR")
        # _NET_WM_STATE_SKIP_PAGER = display.intern_atom(
        #     "_NET_WM_STATE_SKIP_PAGER")
        _NET_WM_STATE_BELOW = display.intern_atom(
            "_NET_WM_STATE_BELOW")
        # _NET_WM_USER_TIME = display.intern_atom(
        #     "_NET_WM_USER_TIME")
        _NET_WM_WINDOW_TYPE = display.intern_atom(
            "_NET_WM_WINDOW_TYPE")
        _NET_WM_WINDOW_TYPE_DOCK = display.intern_atom(
            "_NET_WM_WINDOW_TYPE_DOCK")
        # noinspection PyPep8Naming
        CARDINAL = display.intern_atom("CARDINAL")

        # Set window type to dock (skip task bar, no focus, on all desktops)
        window.change_property(
            _NET_WM_WINDOW_TYPE, ATOM, 32, [_NET_WM_WINDOW_TYPE_DOCK])
        # Reserve space for title bar and task bar
        window.change_property(
            _NET_WM_STRUT, CARDINAL, 32, [0, 0, 28, 28])

        # Show on all workspaces
        # window.change_property(
        #     _NET_WM_DESKTOP, CARDINAL, 32, [0xffffffff, 0, 0, 0])
        # Don't focus the window on window creation
        # window.change_property(
        #     _NET_WM_USER_TIME, CARDINAL, 32, [0, 0, 0, 0])
        # Don't show in task bar or window switcher
        # window.change_property(
        #     _NET_WM_STATE, ATOM, 32,
        #     [_NET_WM_STATE_SKIP_TASKBAR, _NET_WM_STATE_SKIP_PAGER])

        #
        window.change_property(_NET_WM_STATE, ATOM, 32, [_NET_WM_STATE_BELOW])

        display.sync()

    def set_shell_widget(self, shell_widget):
        self.setCentralWidget(shell_widget)
